import enum

from mlflow.entities import ViewType

from mlfoundry.exceptions import MlFoundryException


class ViewType(ViewType):
    """
    Available Keys:
    ACTIVE_ONLY: 1 ,DELETED_ONLY: 2, ALL: 3
    """

    pass


class EnumMissingMixin:
    @classmethod
    def _missing_(cls, value):
        raise MlFoundryException(
            "%r is not a valid %s.  Valid types: %s"
            % (
                value,
                cls.__name__,
                ", ".join([repr(m.value) for m in cls]),
            )
        )


class FileFormat(EnumMissingMixin, enum.Enum):
    CSV = "csv"
    PARQUET = "parquet"


class ModelFramework(EnumMissingMixin, enum.Enum):
    SKLEARN = "sklearn"
    TENSORFLOW = "tensorflow"
    PYTORCH = "pytorch"
    KERAS = "keras"
    XGBOOST = "xgboost"
    LIGHTGBM = "lightgbm"
    FASTAI = "fastai"
    H2O = "h2o"
    ONNX = "onnx"
    SPACY = "spacy"
    STATSMODELS = "statsmodels"
    GLUON = "gluon"
    PADDLE = "paddle"
    TRANSFORMERS = "transformers"
    UNKNOWN = "unknown"


class DataSlice(EnumMissingMixin, enum.Enum):
    TRAIN = "train"
    VALIDATE = "validate"
    TEST = "test"
    PREDICTION = "prediction"


class ModelType(EnumMissingMixin, enum.Enum):
    BINARY_CLASSIFICATION = "binary_classification"
    MULTICLASS_CLASSIFICATION = "multiclass_classification"
    REGRESSION = "regression"
    TIMESERIES = "timeseries"



import os
import posixpath
import tempfile
import time
import typing

from mlflow.entities import RunLog
from mlflow.tracking import MlflowClient
from mlflow.utils.file_utils import download_file_using_http_uri
from pydantic import BaseModel

from mlfoundry.exceptions import MlFoundryException


class PydanticBase(BaseModel):
    # I can make this a property,
    # but <3.9, it is difficult to access
    # property from classmethod
    @staticmethod
    def get_log_type() -> str:
        raise NotImplementedError()

    def to_run_log(
        self, key: str, step: int = 0, timestamp: typing.Optional[int] = None
    ) -> RunLog:
        timestamp = timestamp or int(time.time() * 1000)
        return RunLog(
            key=key,
            step=step,
            timestamp=timestamp,
            log_type=self.get_log_type(),
            value=self.json(),
        )

    def to_run_log_as_artifact(
        self,
        key: str,
        run_id: str,
        mlflow_client: MlflowClient,
        file_name: str,
        artifact_path: str,
        step: int = 0,
        timestamp: typing.Optional[int] = None,
    ) -> RunLog:
        with tempfile.TemporaryDirectory() as local_dir:
            file_path = os.path.join(local_dir, file_name)
            with open(file_path, "w") as fp:
                fp.write(self.json())

            mlflow_client.log_artifact(
                run_id=run_id, local_path=file_path, artifact_path=artifact_path
            )
        artifact_path = posixpath.join(artifact_path, file_name)
        timestamp = timestamp or int(time.time() * 1000)
        return RunLog(
            key=key,
            step=step,
            timestamp=timestamp,
            log_type=self.get_log_type(),
            artifact_path=artifact_path,
        )

    @classmethod
    def from_run_log(cls, run_log: RunLog):
        if run_log.log_type != cls.get_log_type():
            raise MlFoundryException(
                f"run_log.log_type {run_log.log_type} not matching with "
                f"{cls.__name__} log_type {cls.get_log_type()}"
            )
        value = run_log.value
        if value:
            return cls.parse_raw(value)

        artifact_signed_uri = run_log.artifact_signed_uri
        with tempfile.NamedTemporaryFile(mode="w") as file_path:
            download_file_using_http_uri(
                http_uri=artifact_signed_uri, download_path=file_path.name
            )
            return cls.parse_file(file_path.name)



from mlfoundry.monitoring.store.repositories.rest_monitoring_store import (
    RestMonitoringStore,
)
from mlfoundry.session import Session


def get_monitoring_store(session: Session):
    # For now we have only one repo
    return RestMonitoringStore(
        get_host_creds=session.get_monitoring_foundry_host_creds_builder()
    )



import enum
import json
import os
import tempfile
import warnings
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
from transformers import IntervalStrategy, Trainer
from transformers.integrations import rewrite_logs
from transformers.trainer_callback import TrainerCallback
from transformers.utils import flatten_dict

import mlfoundry
from mlfoundry.enums import EnumMissingMixin
from mlfoundry.logger import logger
from mlfoundry.mlfoundry_api import _resolve_ml_repo_name
from mlfoundry.mlfoundry_run import MlFoundryRun

__all__ = ["HF_MODEL_PATH", "MlFoundryTrainerCallback", "LogModelStrategy"]


HF_ARTIFACTS_PATH = "hf/"
HF_MODELS_PATH = os.path.join(HF_ARTIFACTS_PATH, "models")
# intended to be imported while loading back the model
HF_MODEL_PATH = os.path.join(HF_MODELS_PATH, "model")


class LogModelStrategy(str, EnumMissingMixin, enum.Enum):
    NO = "no"
    BEST_ONLY = "best"
    BEST_PLUS_LATEST = "best_plus_latest"
    # ALL_CHECKPOINTS = "all_checkpoints"  # TODO: Implement this later
    CHECKPOINTS_ON_TRAIN_END = "checkpoints_on_train_end"


def _get_artifact_path_for_checkpoint(checkpoint_path):
    return os.path.join(HF_MODELS_PATH, os.path.basename(checkpoint_path))


# TODO: Allow disabling logging via environment variable and configuring settings from env
# TODO: Add support for state.is_hyper_param_search
class MlFoundryTrainerCallback(TrainerCallback):
    """
    Huggingface Transformers Trainer Callback for tracking training run on MLFoundry

    Examples:

        ```
        import mlfoundry
        mlfoundry.login()
        # or set API key via `TFY_API_KEY` env variable

        from transformers import TrainingArguments, Trainer
        from mlfoundry.integrations.transformers import MlFoundryTrainerCallback, LogModelStrategy

        mlf_cb = MlFoundryTrainerCallback(
            project_name="huggingface",
            run_name="my-hf-run",
            flatten_params=True,
            log_model_strategy=LogModelStrategy.BEST_PLUS_LATEST,
        )

        args = TrainingArguments(..., report_to=[])
        trainer = Trainer(..., args=args, callbacks=[mlf_cb])
        trainer.train()
        ```

        Callback can also be created from an existing run

        ```
        import mlfoundry
        mlfoundry.login()
        # or set API key via `TFY_API_KEY` env variable

        from transformers import TrainingArguments, Trainer
        import mlfoundry
        from mlfoundry.integrations.transformers import MlFoundryTrainerCallback, LogModelStrategy

        client = mlfoundry.get_client()
        run = client.create_run(project_name="huggingface", run_name="my-hf-run")

        mlf_cb = MlFoundryTrainerCallback.from_run(
            run=run,
            auto_end_run=False,
            flatten_params=True,
            log_model_strategy=LogModelStrategy.BEST_PLUS_LATEST,
        )

        args = TrainingArguments(..., report_to=[])
        trainer = Trainer(..., args=args, callbacks=[mlf_cb])
        trainer.train()

        run.end()
        ```
    """

    def __init__(
        self,
        ml_repo: Optional[str] = None,
        run_name: Optional[str] = None,
        flatten_params: bool = False,
        log_model_strategy: Union[str, LogModelStrategy] = LogModelStrategy.NO,
        project_name: Optional[str] = None,
        **kwargs,
    ):
        """
        Args:
            ml_repo (str): name of the ML Repo to create the run under
            run_name (Optional[str], optional): name of the run. When not provided a run name is automatically generated
            flatten_params (bool, optional): if to flatten the args and model config dictionaries before logging them,
                By default, this is `False`

                For e.g.

                when set to True,
                `config = {"id2label": {"0": "hello", "1": "bye"}}` will be logged as two parameters as
                `{"config/id2label.0": "hello", "config/id2label.1": "bye"}`

                when set to False,
                `config = {"id2label": {"0": "hello"}}` will be logged as a single parameter as
                `{"config/id2label": '{"0": "hello", "1": "bye"}'}`

            log_model_strategy (LogModelStrategy, optional): The strategy to use for logging models
                - LogModelStrategy.NO (default): Do not log any models
                - LogModelStrategy.BEST_ONLY: Log only the best model checkpoint.
                    Note that `args.metric_for_best_model` and `args.args.save_strategy` needs to be set
                    correctly for this to work
                - LogModelStrategy.BEST_PLUS_LATEST: Log both the latest checkpoint and the best checkpoint (if
                    available) and different from the latest checkpoint
                - LogModelStrategy.CHECKPOINTS_ON_TRAIN_END: Log all available checkpoints in `args.output_dir` at the
                    end of training.

                In addition to checkpoints a metadata.json file is logged which stores references to the best checkpoint
                and the latest checkpoint

                The files will be logged under `mlf/huggingface_models/model/` under the run artifacts.
            log_model (bool, optional): **DEPRECATED**, use `log_model_strategy` instead.
                If to log the generated model artifacts at the end of training. By default, this is `False`.
                If set to True this will set `log_model_strategy` to `LogModelStrategy.BEST_PLUS_LATEST`. Please
                see the description for `log_model_strategy`
        """
        log_model_strategy = LogModelStrategy(log_model_strategy)
        if "log_model" in kwargs:
            _option = (
                LogModelStrategy.BEST_PLUS_LATEST.value
                if kwargs["log_model"]
                else LogModelStrategy.NO.value
            )
            warnings.warn(
                f"`log_model` argument has been deprecated, please use the `log_model_strategy` argument instead. "
                f"E.g. log_model_strategy={_option}",
                FutureWarning,
            )
            if kwargs["log_model"] and log_model_strategy == LogModelStrategy.NO:
                log_model_strategy = LogModelStrategy.BEST_PLUS_LATEST

        self._ml_repo = _resolve_ml_repo_name(
            ml_repo=ml_repo, project_name=project_name
        )
        self._run_name = run_name
        self._run: Optional[MlFoundryRun] = None
        self._auto_end_run = True

        self._flatten_params = flatten_params
        self._log_model_strategy = log_model_strategy
        self._MAX_PARAM_VAL_LENGTH = 250
        self._MAX_PARAMS_TAGS_PER_BATCH = 100
        self._initialized = False
        self._last_save_step = -1

    @classmethod
    def from_run(
        cls,
        run: MlFoundryRun,
        auto_end_run: bool = False,
        flatten_params: bool = False,
        log_model_strategy: Union[str, LogModelStrategy] = LogModelStrategy.NO,
        **kwargs,
    ) -> "MlFoundryTrainerCallback":
        """
        Create a MLFoundry Huggingface Transformers Trainer callback from an existing MLFoundry run instance

        Args:
            run (MlFoundryRun): `MlFoundry` run instance
            auto_end_run (bool, optional): if to end the run when training finishes. By default, this is `False`
            flatten_params (bool, optional): if to flatten the args and model config dictionaries before logging them,
                By default, this is `False`

                For e.g.

                when set to True,
                `config = {"id2label": {"0": "hello", "1": "bye"}}` will be logged as two parameters as
                `{"config/id2label.0": "hello", "config/id2label.1": "bye"}`

                when set to False,
                `config = {"id2label": {"0": "hello"}}` will be logged as a single parameter as
                `{"config/id2label": '{"0": "hello", "1": "bye"}'}`

            log_model_strategy (LogModelStrategy, optional): The strategy to use for logging models
                - LogModelStrategy.NO (default): Do not log any models
                - LogModelStrategy.BEST_ONLY: Log only the best model checkpoint.
                    Note that `args.metric_for_best_model` and `args.args.save_strategy` needs to be set
                    correctly for this to work
                - LogModelStrategy.BEST_PLUS_LATEST: Log both the latest checkpoint and the best checkpoint (if
                    available) and different from the latest checkpoint
                - LogModelStrategy.CHECKPOINTS_ON_TRAIN_END: Log all available checkpoints in `args.output_dir` at the
                    end of training.

                In addition to checkpoints a metadata.json file is logged which stores references to the best checkpoint
                and the latest checkpoint

                The files will be logged under `mlf/huggingface_models/model/` under the run artifacts.
            log_model (bool, optional): **DEPRECATED**, use `log_model_strategy` instead.
                If to log the generated model artifacts at the end of training. By default, this is `False`.
                If set to True this will set `log_model_strategy` to `LogModelStrategy.BEST_PLUS_LATEST`. Please
                see the description for `log_model_strategy`
            **kwargs: Additional keyword arguments to pass to init

        Returns:
            MlFoundryTrainerCallback: an instance of `MlFoundryTrainerCallback`
        """
        instance = cls(
            ml_repo=run.ml_repo,
            run_name=run.run_name,
            flatten_params=flatten_params,
            log_model_strategy=log_model_strategy,
            **kwargs,
        )
        instance._run = run
        instance._auto_end_run = auto_end_run
        return instance

    def on_init_end(self, args, state, control, **kwargs):
        """
        Event called at the end of the initialization of the [`Trainer`].
        """
        if self._log_model_strategy == LogModelStrategy.BEST_ONLY:
            if args.save_strategy == IntervalStrategy.NO:
                logger.warning(
                    f"`log_model_strategy` is set to BEST_ONLY but "
                    f"`args.save_strategy` is set to NO which means best model cannot be tracked automatically. "
                    f"Falling back to BEST_PLUS_LATEST."
                )
                self._log_model_strategy = LogModelStrategy.BEST_PLUS_LATEST
            elif args.metric_for_best_model is None:
                logger.warning(
                    f"`log_model_strategy` is set to BEST_ONLY but "
                    f"`args.metric_for_best_model` is set to None "
                    f"which means best model cannot be tracked automatically. "
                    f"Falling back to BEST_PLUS_LATEST."
                )
                self._log_model_strategy = LogModelStrategy.BEST_PLUS_LATEST
            # TODO (chiragjn): Check for a case when save steps is lower than save frequency

        if self._log_model_strategy == LogModelStrategy.BEST_PLUS_LATEST:
            if args.save_strategy == IntervalStrategy.NO:
                logger.warning(
                    f"`args.save_strategy` is set to NO "
                    f"which means best model cannot be tracked automatically. "
                    f"Only latest model will be saved."
                )
            elif args.metric_for_best_model is None:
                logger.warning(
                    f"`args.metric_for_best_model` is set to None "
                    f"which means best model cannot be tracked automatically. "
                    f"automatically. Only latest model will be saved."
                )

    def setup(self, args, state, model, **kwargs):
        if not state.is_world_process_zero:
            # If the current process is not the global main process in a distributed training setting do nothing
            return

        logger.info("Automatic MLFoundry logging enabled")

        if not self._run:
            self._auto_end_run = True
            self._run = mlfoundry.get_client().create_run(
                ml_repo=self._ml_repo,
                run_name=self._run_name,
            )

        args_dict = {f"args/{k}": v for k, v in args.to_dict().items()}
        if self._flatten_params:
            args_dict = flatten_dict(args_dict)

        if hasattr(model, "config") and model.config is not None:
            model_config_dict = {
                f"config/{k}": v for k, v in model.config.to_dict().items()
            }
            if self._flatten_params:
                model_config_dict = flatten_dict(model_config_dict)
        else:
            model_config_dict = {}

        params: Dict[str, Any] = {**args_dict, **model_config_dict}
        for name, value in list(params.items()):
            # internally, all values are converted to str
            if len(str(value)) > self._MAX_PARAM_VAL_LENGTH:
                logger.warning(
                    f'Trainer is attempting to log a value of "{value}" for key "{name}" as a parameter. '
                    f"MlFoundry's log_params() only accepts values no longer than {self._MAX_PARAM_VAL_LENGTH} "
                    f"characters so we dropped this attribute. Pass `flatten_params=True` during init to "
                    f"flatten the parameters and avoid this message."
                )
                del params[name]

        # MlFoundry cannot log more than 100 values in one go, so we have to split it
        params_items: List[Tuple[str, Any]] = list(params.items())
        for i in range(0, len(params_items), self._MAX_PARAMS_TAGS_PER_BATCH):
            self._run.log_params(
                dict(params_items[i : i + self._MAX_PARAMS_TAGS_PER_BATCH])
            )
        self._initialized = True

    def on_train_begin(self, args, state, control, model=None, **kwargs):
        """
        Event called at the beginning of training.
        """
        if not self._initialized:
            self.setup(args, state, model)

    def _get_latest_checkpoint_path(
        self,
        temp_dir,
        args,
        state,
        model=None,
        tokenizer=None,
        optimizer=None,
        lr_scheduler=None,
        **kwargs,
    ) -> Optional[str]:
        latest_checkpoint_path = None
        expected_latest_checkpoint_name = f"checkpoint-{state.global_step}"
        expected_latest_checkpoint_path = os.path.join(
            args.output_dir, expected_latest_checkpoint_name
        )

        # first check if a checkpoint for latest global step exists on disk
        if os.path.exists(expected_latest_checkpoint_path):
            latest_checkpoint_path = expected_latest_checkpoint_path
        # else check if current `model` is not (re)loaded from the best checkpoint
        # (although best cloud possibly be the latest too), then save it
        elif not args.load_best_model_at_end or state.best_model_checkpoint is None:
            latest_checkpoint_path = os.path.join(
                temp_dir, expected_latest_checkpoint_name
            )
            fake_trainer = Trainer(
                args=args,
                model=model,
                tokenizer=tokenizer,
                optimizers=(optimizer, lr_scheduler),
            )
            fake_trainer.save_model(latest_checkpoint_path)
        # else fallback to the last known save step to be the latest
        elif self._last_save_step > -1:
            # TODO: Check if there is a possibility that this gets deleted in rotation logic
            expected_latest_checkpoint_path = os.path.join(
                args.output_dir, f"checkpoint-{self._last_save_step}"
            )
            if os.path.exists(expected_latest_checkpoint_path):
                latest_checkpoint_path = expected_latest_checkpoint_path

        return latest_checkpoint_path

    def _get_best_checkpoint_path(self, state) -> Optional[str]:
        best_checkpoint_path = None
        if state.best_model_checkpoint is not None and os.path.exists(
            state.best_model_checkpoint
        ):
            best_checkpoint_path = state.best_model_checkpoint
        return best_checkpoint_path

    def _legacy_log_model(
        self,
        temp_dir: str,
        best_checkpoint_path: Optional[str],
        latest_checkpoint_path: Optional[str],
    ):
        # TODO: this is kept only for backward compatibility and duplicates a checkpoint
        #       this should be removed
        checkpoint_path = best_checkpoint_path or latest_checkpoint_path
        if not checkpoint_path:
            return

        self._run.log_artifact_deprecated(
            local_path=checkpoint_path, artifact_path=HF_MODEL_PATH
        )
        metadata = {"checkpoint_name": os.path.basename(checkpoint_path)}
        metadata_filepath = os.path.join(temp_dir, ".mlfoundry.json")
        with open(metadata_filepath, "w") as f:
            json.dump(metadata, f)
        self._run.log_artifact_deprecated(
            local_path=metadata_filepath, artifact_path=HF_MODEL_PATH
        )

    def _log_model_on_train_end(
        self,
        temp_dir,
        args,
        state,
        control,
        **kwargs,
    ):
        metadata = {"checkpoints": {}}
        latest_checkpoint_path = self._get_latest_checkpoint_path(
            temp_dir=temp_dir, args=args, state=state, **kwargs
        )
        latest_checkpoint_name = (
            os.path.basename(latest_checkpoint_path) if latest_checkpoint_path else None
        )
        best_checkpoint_path = self._get_best_checkpoint_path(state=state)
        best_checkpoint_name = (
            os.path.basename(best_checkpoint_path) if best_checkpoint_path else None
        )

        checkpoints_to_log: List[str] = []
        if self._log_model_strategy == LogModelStrategy.BEST_ONLY:
            latest_checkpoint_path = None
            if best_checkpoint_path is not None:
                checkpoints_to_log.append(best_checkpoint_path)
            else:
                logger.warning(
                    f"`log_model_strategy` was set to BEST_ONLY but Trainer did not save any best "
                    f"checkpoint. Cannot automatically log model"
                )
        elif self._log_model_strategy == LogModelStrategy.CHECKPOINTS_ON_TRAIN_END:
            # Log all "checkpoint-\d+" in args.output_dir
            output_dir_contents = os.listdir(args.output_dir)
            for file_or_dir in output_dir_contents:
                if os.path.isdir(file_or_dir) and file_or_dir.startswith("checkpoint-"):
                    checkpoint_path = os.path.join(args.output_dir, file_or_dir)
                    checkpoints_to_log.append(checkpoint_path)
            if (
                latest_checkpoint_path
                and latest_checkpoint_name not in output_dir_contents
            ):
                # it is possible that state.global_step > self._last_save_step
                checkpoints_to_log.append(latest_checkpoint_path)
        elif self._log_model_strategy == LogModelStrategy.BEST_PLUS_LATEST:
            if latest_checkpoint_path:
                checkpoints_to_log.append(latest_checkpoint_path)
            if best_checkpoint_path and best_checkpoint_name != latest_checkpoint_name:
                checkpoints_to_log.append(best_checkpoint_path)
        else:
            raise NotImplementedError("Unreachable!")

        for checkpoint_path in checkpoints_to_log:
            checkpoint_name = os.path.basename(checkpoint_path)
            artifact_path = os.path.join(HF_MODELS_PATH, checkpoint_name)
            logger.debug(
                f"Logging checkpoint at {checkpoint_path!r} to {artifact_path!r}"
            )
            self._run.log_artifact_deprecated(
                local_path=checkpoint_path, artifact_path=artifact_path
            )

        if best_checkpoint_path:
            metadata["checkpoints"]["best"] = _get_artifact_path_for_checkpoint(
                best_checkpoint_path
            )

        if latest_checkpoint_path:
            metadata["checkpoints"]["latest"] = _get_artifact_path_for_checkpoint(
                latest_checkpoint_path
            )

        metadata_file_name = "metadata.json"
        metadata_path = os.path.join(temp_dir, metadata_file_name)
        with open(metadata_path, "w") as f:
            json.dump(metadata, f)
        self._run.log_artifact_deprecated(
            local_path=metadata_path,
            artifact_path=HF_MODELS_PATH,
        )

        self._legacy_log_model(
            temp_dir=temp_dir,
            best_checkpoint_path=best_checkpoint_path,
            latest_checkpoint_path=latest_checkpoint_path,
        )

    def on_train_end(self, args, state, control, **kwargs):
        """
        Event called at the end of training.
        """
        if not self._initialized or not state.is_world_process_zero:
            return
        if self._log_model_strategy != LogModelStrategy.NO:
            logger.info("Logging artifacts. This may take time.")
            with tempfile.TemporaryDirectory() as temp_dir:
                self._log_model_on_train_end(
                    temp_dir=temp_dir, args=args, state=state, control=control, **kwargs
                )
        if self._auto_end_run and self._run:
            self._run.end()

    def on_log(self, args, state, control, model=None, logs=None, **kwargs):
        """
        Event called after logging the last logs.
        """
        if not self._initialized:
            self.setup(args, state, model)
        if not state.is_world_process_zero:
            return
        logs = rewrite_logs(logs)
        metrics = {}
        for k, v in logs.items():
            if isinstance(v, (int, float, np.integer, np.floating)):
                metrics[k] = v
            else:
                logger.warning(
                    f'Trainer is attempting to log a value of "{v}" of type {type(v)} for key "{k}" as a metric. '
                    f"MlFoundry's log_metric() only accepts float and int types so we dropped this attribute."
                )
        self._run.log_metrics(metric_dict=metrics, step=state.global_step)

    def on_save(self, args, state, control, **kwargs):
        self._last_save_step = state.global_step

    def __del__(self):
        if self._auto_end_run and self._run:
            self._run.end()



import collections
import json
import os.path
import tempfile
from typing import Any, Dict, List, Optional, Tuple, Union

from mlflow.entities import ArtifactType
from mlflow.tracking import MlflowClient

from mlfoundry.exceptions import MlFoundryException
from mlfoundry.log_types.artifacts.artifact import (
    ArtifactPath,
    ArtifactVersion,
    ArtifactVersionInternalMetadata,
    _log_artifact_version_helper,
)
from mlfoundry.log_types.artifacts.constants import FILES_DIR, INTERNAL_METADATA_PATH
from mlfoundry.log_types.artifacts.utils import (
    _copy_additional_files,
    _validate_artifact_metadata,
    _validate_description,
)
from mlfoundry.logger import logger


def _log_artifact_version(
    run,
    name: str,
    artifact_paths: List[Union[ArtifactPath, Tuple[str, Optional[str]]]],
    mlflow_client: Optional[MlflowClient] = None,
    ml_repo_id: Optional[str] = None,
    description: Optional[str] = None,
    metadata: Optional[Dict[str, Any]] = None,
    step: int = 0,
) -> ArtifactVersion:

    for i, artifact_path in enumerate(artifact_paths):
        if isinstance(artifact_path, ArtifactPath):
            continue
        elif isinstance(artifact_path, collections.abc.Sequence) and (
            0 < len(artifact_path) <= 2
        ):
            artifact_paths[i] = ArtifactPath(*artifact_path)
        else:
            raise ValueError(
                "`artifact_path` should be an instance of `mlfoundry.ArtifactPath` or a tuple of (src, dest) path strings"
            )

    metadata = metadata or {}
    step = step or 0

    _validate_description(description)
    _validate_artifact_metadata(metadata)

    logger.info("Logging the artifact, this might take a while ...")
    temp_dir = tempfile.TemporaryDirectory(prefix="truefoundry-")

    internal_metadata = ArtifactVersionInternalMetadata(
        files_dir=FILES_DIR,
    )

    try:
        local_files_dir = os.path.join(temp_dir.name, internal_metadata.files_dir)
        os.makedirs(local_files_dir, exist_ok=True)

        logger.info("Copying the files to log")
        _copy_additional_files(
            root_dir=temp_dir.name,
            files_dir=internal_metadata.files_dir,
            model_dir=None,
            additional_files=artifact_paths,
        )

    # TODO(nikp1172) verify error messag when artifact doesn't exist
    except Exception as e:
        temp_dir.cleanup()
        raise MlFoundryException("Failed to log artifact") from e

    # save internal metadata
    local_internal_metadata_path = os.path.join(temp_dir.name, INTERNAL_METADATA_PATH)
    os.makedirs(os.path.dirname(local_internal_metadata_path), exist_ok=True)
    with open(local_internal_metadata_path, "w") as f:
        json.dump(internal_metadata.dict(), f)

    return _log_artifact_version_helper(
        run=run,
        mlflow_client=mlflow_client,
        ml_repo_id=ml_repo_id,
        name=name,
        artifact_type=ArtifactType.ARTIFACT,
        artifact_dir=temp_dir,
        description=description,
        metadata=metadata,
        step=step,
    )



"""
Define all the env variable names that users
can use
"""

# NOTE: These are shared between servicefoundry and mlfoundry
_TRACKING_HOST_GLOBAL = "TFY_HOST"
API_KEY_GLOBAL = "TFY_API_KEY"
DISABLE_MULTIPART_UPLOAD = "MLF_DISABLE_MULTIPART_UPLOAD"


TRACKING_TOKEN = "MLFLOW_TRACKING_TOKEN"

INTERNAL_ENV_VARS = [
    "TFY_INTERNAL_APPLICATION_ID",
    "TFY_INTERNAL_JOB_RUN_NAME",
]



from importlib_metadata import version

# catching any exception to make sure
# the `import mlfoundry` code does not fail
# if `version` function fails for some reason
try:
    __version__ = version("mlfoundry")
except Exception:
    __version__ = None



from datetime import datetime
from typing import Dict, List, Optional, Union

from pydantic import BaseModel, Field, constr

# TODO @nikp1172 Add support for np types https://numpy.org/doc/stable/user/basics.types.html
EntityValue = Union[str, int, float]


class PredictionData(BaseModel):
    value: EntityValue
    probabilities: Dict[str, float] = Field(default_factory=dict)
    shap_values: Dict[str, float] = Field(default_factory=dict)

    class Config:
        smart_union = True


class BasePacket(BaseModel):
    class Config:
        smart_union = True


class Prediction(BaseModel):

    data_id: constr(min_length=1, max_length=64)
    features: Dict[str, EntityValue]
    prediction_data: PredictionData
    actual_value: Optional[EntityValue] = None

    raw_data: Optional[Dict] = Field(default_factory=dict)
    occurred_at: datetime = Field(
        default_factory=lambda: datetime.utcnow().replace(tzinfo=None)
    )

    class Config:
        smart_union = True


class PredictionPacket(BasePacket):
    model_version_id: constr(min_length=1, max_length=64)
    prediction: Prediction


class Actual(BaseModel):
    data_id: constr(min_length=1, max_length=64)
    value: EntityValue

    class Config:
        smart_union = True


class ActualPacket(BasePacket):
    model_version_id: constr(min_length=1, max_length=64)
    actual: Actual


class ModelVersion(BaseModel):
    class Config:
        use_enum_values = True
        smart_union = True

    id: str
    fqn: str
    model_id: str
    model_fqn: str
    version: int

    model_schema: Optional[Dict] = None
    custom_metrics: Optional[List[Dict]] = Field(default_factory=list)
    created_by: str
    created_at: datetime
    updated_at: datetime
    monitoring_enabled: bool


# Kept only the fields relevant for the user
class DatasetData(BaseModel):
    data_id: str
    features: Dict[str, EntityValue]
    actual: Dict[str, Optional[EntityValue]]
    raw_data: Dict
    created_at: Optional[datetime]



"""
# TO independently test this module, you can run the example in the path
python examples/sklearn/iris_train.py

Besides running pytest
"""
from __future__ import annotations

import os
import warnings
from pathlib import Path
from typing import (
    Any,
    Dict,
    Generator,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import coolname
import mlflow
import pandas as pd
from mlflow.entities import Artifact, ArtifactType, CustomMetric, Model, ModelSchema
from mlflow.store.tracking import _SEARCH_MAX_RESULTS_DEFAULT
from mlflow.tracking import MlflowClient

from mlfoundry import amplitude, constants, env_vars
from mlfoundry.enums import ModelFramework, ViewType
from mlfoundry.env_vars import _TRACKING_HOST_GLOBAL
from mlfoundry.exceptions import MlflowException, MlFoundryException
from mlfoundry.internal_namespace import NAMESPACE
from mlfoundry.log_types.artifacts.artifact import ArtifactPath, ArtifactVersion
from mlfoundry.log_types.artifacts.general_artifact import _log_artifact_version
from mlfoundry.log_types.artifacts.model import ModelVersion, _log_model_version
from mlfoundry.logger import logger
from mlfoundry.mlfoundry_run import MlFoundryRun
from mlfoundry.monitoring.store import MonitoringClient
from mlfoundry.session import Session, get_active_session, init_session
from mlfoundry.tracking.servicefoundry_service import ServicefoundryService


def _get_internal_env_vars_values() -> Dict[str, str]:
    env = {}
    for env_var_name in env_vars.INTERNAL_ENV_VARS:
        value = os.getenv(env_var_name)
        if value:
            env[env_var_name] = value

    return env


def get_client(disable_analytics: bool = False) -> MlFoundry:
    """Initializes and returns the mlfoundry client.

    Args:
        disable_analytics (bool, optional): To turn off usage analytics collection, pass `True`.
            By default, this is set to `False`.

    Returns:
        MlFoundry: Instance of `MlFoundry` class which represents a `run`.

    Examples:
        ### Get client
        ```python
        import mlfoundry

        client = mlfoundry.get_client()
        ```
    """
    user_id = amplitude.NO_USER

    session = None

    # NOTE: hack to run tests
    if os.getenv(_TRACKING_HOST_GLOBAL, "").startswith("file:"):
        tracking_uri = os.getenv(_TRACKING_HOST_GLOBAL)
        tracking_uri = os.path.join(tracking_uri, constants.MLRUNS_FOLDER_NAME)
        mlflow.set_tracking_uri(tracking_uri)
    else:
        session = init_session()
        user_id = session.user_info.user_id

    amplitude.init(user_id=user_id, disable_analytics=disable_analytics)
    amplitude.track(amplitude.Event.GET_CLIENT)
    return MlFoundry(session=session)


def _resolve_ml_repo_name(
    ml_repo: Optional[str] = None,
    project_name: Optional[str] = None,
) -> str:
    if project_name and ml_repo:
        raise MlFoundryException(
            f"Only one of `ml_repo` or `project_name` should be passed"
        )
    if not project_name and not ml_repo:
        raise MlFoundryException(f"`ml_repo` must be string type and cannot be empty")
    if project_name:
        warning_message = (
            "Argument `project_name` has been deprecated and renamed to `ml_repo`. "
            "Please use `ml_repo` as `project_name` will be removed in the upcoming versions"
        )
        warnings.warn(warning_message, FutureWarning)
        logger.warning(warning_message)
    if ml_repo:
        project_name = ml_repo
    if project_name == "" or (not isinstance(project_name, str)):
        raise MlFoundryException(
            f"ml_repo must be string type and not empty. "
            f"Got {type(project_name)} type with value {project_name!r}"
        )
    return project_name


def _resolve_version(version: Union[int, str]) -> int:
    if not isinstance(version, int) and not (
        isinstance(version, str) and version.isnumeric()
    ):
        raise MlFoundryException(
            f"version must be an integer or string containing numbers only. Got {version!r}"
        )
    final_version = int(version)
    if final_version <= 0:
        raise ValueError("version must be greater than 0")
    return final_version


class MlFoundry:
    """MlFoundry."""

    def __init__(self, session: Optional[Session] = None):
        """__init__.

        Args:
            session (Optional[Session], optional): Session instance to get auth credentials from
        """
        self.mlflow_client = MlflowClient()
        if session:
            self.monitoring_client = MonitoringClient(session=session)

    def _get_ml_repo(self, ml_repo: str) -> str:
        """_get_ml_repo.

        Args:
            ml_repo (str): The name of the ML Repo.

        Returns:
            str: The id of the ML Repo.
        """
        try:
            ml_repo_obj = self.mlflow_client.get_experiment_by_name(name=ml_repo)
        except MlflowException as e:
            err_msg = (
                f"Error happened in getting ML Repo based on name: "
                f"{ml_repo}. Error details: {e.message}"
            )
            raise MlFoundryException(err_msg) from e
        if not ml_repo_obj:
            err_msg = (
                f"ML Repo Does Not Exist for name: {ml_repo}. You may either "
                "create it from the dashboard or using client.create_ml_repo('<ml_repo_name>')"
            )
            raise MlFoundryException(err_msg)
        return ml_repo_obj.experiment_id

    def list_ml_repos(self) -> List[str]:
        """Returns a list of names of ML Repos accessible by the current user.

        Returns:
            List[str]: A list of names of ML Repos
        """
        amplitude.track(amplitude.Event.GET_ALL_PROJECTS)
        try:
            ml_repos = self.mlflow_client.list_experiments(view_type=ViewType.ALL)
        except MlflowException as e:
            err_msg = f"Error happened in fetching ML Repos. Error details: {e.message}"
            raise MlFoundryException(err_msg) from e

        ml_repos_names = []
        for ml_repo in ml_repos:
            # ML Repo with experiment_id 0 represents default ML Repo which we are removing.
            if ml_repo.experiment_id != "0":
                ml_repos_names.append(ml_repo.name)

        return ml_repos_names

    def create_ml_repo(
        self,
        ml_repo: Optional[str] = None,
        storage_integration_fqn: Optional[str] = None,
    ):
        """Creates an ML Repository.

        Args:
            ml_repo (str, optional): The name of the Repository you want to create.
                if not given, it creates a name by itself.

            storage_integration_fqn(str, optional): The storage integration FQN to use for the experiment for saving artifacts.

        Examples:
                ### Create Repository
                ```python
                import mlfoundry

                client = mlfoundry.get_client()

                ml_repo = client.create_ml_repo(ml_repo="my-repo")
            ```
        """
        existing_ml_repo = self.mlflow_client.get_experiment_by_name(name=ml_repo)
        if not existing_ml_repo:  # ml_repo does not exist
            self.mlflow_client.get_experiment(
                experiment_id=self.mlflow_client.create_experiment(
                    name=ml_repo, storage_integration_fqn=storage_integration_fqn
                )
            )

        if storage_integration_fqn:
            session = get_active_session()
            servicefoundry_service = ServicefoundryService(
                tracking_uri=self.get_tracking_uri(),
                token=session.token.access_token,
            )

            try:
                existing_storage_integration = (
                    servicefoundry_service.get_integration_from_id(
                        existing_ml_repo.storage_integration_id
                    )
                )
            except Exception as e:
                err_msg = f"Error happened in getting integration of ML Repo. Error details: {e.message}"
                raise MlFoundryException(err_msg) from e

            if existing_storage_integration["fqn"] != storage_integration_fqn:
                raise MlFoundryException(
                    f"ML Repo with same name already exists with storage integration:"
                    f"{existing_storage_integration['fqn']}. Cannot update the storage integration to: "
                    f"{storage_integration_fqn}"
                )

    def create_run(
        self,
        ml_repo: Optional[str] = None,
        run_name: Optional[str] = None,
        tags: Optional[Dict[str, Any]] = None,
        project_name: Optional[str] = None,
        **kwargs,
    ) -> MlFoundryRun:
        """Initialize a `run`.

        In a machine learning experiment `run` represents a single experiment
        conducted under a project.
        Args:
            ml_repo (str): The name of the project under which the run will be created.
                ml_repo should only contain alphanumerics (a-z,A-Z,0-9) or hyphen (-).
                The user must have `ADMIN` or `WRITE` access to this project.
            run_name (Optional[str], optional): The name of the run. If not passed, a randomly
                generated name is assigned to the run. Under a project, all runs should have
                a unique name. If the passed `run_name` is already used under a project, the
                `run_name` will be de-duplicated by adding a suffix.
                run name should only contain alphanumerics (a-z,A-Z,0-9) or hyphen (-).
            tags (Optional[Dict[str, Any]], optional): Optional tags to attach with
                this run. Tags are key-value pairs.
            kwargs:

        Returns:
            MlFoundryRun: An instance of `MlFoundryRun` class which represents a `run`.

        Examples:
            ### Create a run under current user.
            ```python
            import mlfoundry

            client = mlfoundry.get_client()

            tags = {"model_type": "svm"}
            run = client.create_run(
                ml_repo="my-classification-project", run_name="svm-with-rbf-kernel", tags=tags
            )

            run.end()
            ```

            ### Creating a run using context manager.
            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            with client.create_run(
                ml_repo="my-classification-project", run_name="svm-with-rbf-kernel"
            ) as run:
                # ...
                # Model training code
                ...
            # `run` will be automatically marked as `FINISHED` or `FAILED`.
            ```

            ### Create a run in a project owned by a different user.
            ```python
            import mlfoundry

            client = mlfoundry.get_client()

            tags = {"model_type": "svm"}
            run = client.create_run(
                ml_repo="my-classification-project",
                run_name="svm-with-rbf-kernel",
                tags=tags,
            )
            run.end()
            ```
        """
        amplitude.track(amplitude.Event.CREATE_RUN)

        if not run_name:
            run_name = coolname.generate_slug(2)
            logger.info(
                f"No run_name given. Using a randomly generated name {run_name}."
                " You can pass your own using the `run_name` argument"
            )
        ml_repo = _resolve_ml_repo_name(ml_repo=ml_repo, project_name=project_name)

        ml_repo_id = self._get_ml_repo(ml_repo=ml_repo)

        if tags is not None:
            NAMESPACE.validate_namespace_not_used(tags.keys())
        else:
            tags = {}

        tags.update(_get_internal_env_vars_values())
        run = self.mlflow_client.create_run(ml_repo_id, name=run_name, tags=tags)
        mlf_run_id = run.info.run_id

        mlf_run = MlFoundryRun(
            experiment_id=ml_repo_id, run_id=mlf_run_id, auto_end=True, **kwargs
        )
        # TODO(Rizwan): Revisit this once run lifecycle is formalised
        mlf_run._add_git_info()
        mlf_run._add_python_mlf_version()
        logger.info(f"Run {run.info.fqn!r} has started.")
        print(f"Link to the dashboard for the run: {mlf_run.dashboard_link}")
        return mlf_run

    def get_run_by_id(self, run_id: str) -> MlFoundryRun:
        """Get an existing `run` by the `run_id`.

        Args:
            run_id (str): run_id or fqn of an existing `run`.

        Returns:
            MlFoundryRun: An instance of `MlFoundryRun` class which represents a `run`.

        Examples:

            ### Get run by the run id
            ```python
            import mlfoundry

            client = mlfoundry.get_client()

            run = client.get_run_by_id(run_id='a8f6dafd70aa4baf9437a33c52d7ee90')
            ```

        """
        amplitude.track(amplitude.Event.GET_RUN)
        if run_id == "" or (not isinstance(run_id, str)):
            raise MlFoundryException(
                f"run_id must be string type and not empty. "
                f"Got {type(run_id)} type with value {run_id}"
            )
        if "/" in run_id:
            return self.get_run_by_fqn(run_id)

        run = self.mlflow_client.get_run(run_id)
        mlfoundry_run = MlFoundryRun._get_run_from_mlflow_run(run)
        print(f"Link to the dashboard for the run: {mlfoundry_run.dashboard_link}")
        return mlfoundry_run

    def get_run_by_fqn(self, run_fqn: str) -> MlFoundryRun:
        """Get an existing `run` by `fqn`.

        `fqn` stands for Fully Qualified Name. A run `fqn` has the following pattern:
        tenant_name/ml_repo/run_name

        If  a run `svm` under the project `cat-classifier` in `truefoundry` tenant,
        the `fqn` will be `truefoundry/cat-classifier/svm`.

        Args:
            run_fqn (str): `fqn` of an existing run.

        Returns:
            MlFoundryRun: An instance of `MlFoundryRun` class which represents a `run`.

        Examples:
            ### get run by run fqn
            ```python
            import mlfoundry

            client = mlfoundry.get_client()

            run = client.get_run_by_fqn(run_fqn='truefoundry/my-repo/svm')
            ```

        """
        run = self.mlflow_client.get_run_by_fqn(run_fqn)
        mlfoundry_run = MlFoundryRun._get_run_from_mlflow_run(run)
        print(f"Link to the dashboard for the run: {mlfoundry_run.dashboard_link}")
        return mlfoundry_run

    def get_run_by_name(
        self,
        ml_repo: str,
        run_name: str,
    ) -> MlFoundryRun:
        """Get an existing `run` by `run_name`.

        Args:
            ml_repo (str): name of an the project of which the run is part of.
            run_name (str): the name of the run required

        Returns:
            MlFoundryRun: An instance of `MlFoundryRun` class which represents a `run`.

        Examples:
            ### get run by run run name
            ```python
            import mlfoundry

            client = mlfoundry.get_client()

            run = client.get_run_by_name(run_name='svm', ml_repo='my-repo')
            ```
        """

        run = self.mlflow_client.get_run_by_name(
            experiment_id=None,
            run_name=run_name,
            experiment_name=ml_repo,
        )
        mlfoundry_run = MlFoundryRun._get_run_from_mlflow_run(run)
        print(f"Link to the dashboard for the run: {mlfoundry_run.dashboard_link}")
        return mlfoundry_run

    def get_all_runs(
        self,
        ml_repo: Optional[str] = None,
        project_name: Optional[str] = None,
    ) -> pd.DataFrame:
        """Returns all the run name and id present under a project.

        The user must have `READ` access to the project.
        Args:
            ml_repo (str): Name of the project.
        Returns:
            pd.DataFrame: dataframe with two columns- run_id and run_name

        Examples:
            ### get all the runs from an ml_repo
            ```python
            import mlfoundry

            client = mlfoundry.get_client()

            run = client.get_all_runs(ml_repo='my-repo')
            ```
        """
        amplitude.track(amplitude.Event.GET_ALL_RUNS)
        ml_repo = _resolve_ml_repo_name(ml_repo=ml_repo, project_name=project_name)
        ml_repo_obj = self.mlflow_client.get_experiment_by_name(ml_repo)
        if ml_repo_obj is None:
            return pd.DataFrame(
                columns=[constants.RUN_ID_COL_NAME, constants.RUN_NAME_COL_NAME]
            )

        ml_repo_id = ml_repo_obj.experiment_id

        try:
            all_run_infos = self.mlflow_client.list_run_infos(
                ml_repo_id, run_view_type=ViewType.ALL
            )
        except MlflowException as e:
            err_msg = f"Error happened in while fetching runs for ML Repo {ml_repo}. Error details: {e.message}"
            raise MlFoundryException(err_msg) from e

        runs = []

        for run_info in all_run_infos:
            try:
                run = self.mlflow_client.get_run(run_info.run_id)
                run_name = run.info.name or run.data.tags.get(
                    constants.RUN_NAME_COL_NAME, ""
                )
                runs.append((run_info.run_id, run_name))
            except MlflowException as e:
                logger.warning(
                    f"Could not fetch details of run with run_id {run_info.run_id}. "
                    f"Skipping this one. Error details: {e.message}. "
                )

        return pd.DataFrame(
            runs, columns=[constants.RUN_ID_COL_NAME, constants.RUN_NAME_COL_NAME]
        )

    def search_runs(
        self,
        ml_repo: Optional[str] = None,
        filter_string: str = "",
        run_view_type: str = "ACTIVE_ONLY",
        order_by: Sequence[str] = ("attribute.start_time DESC",),
        project_name: Optional[str] = None,
        job_run_name: Optional[str] = None,
        max_results: Optional[int] = None,
    ) -> Generator[MlFoundryRun, None, None]:
        """
        The user must have `READ` access to the project.
        Returns a Generator that returns a MLFoundryRun on each next call.
        All the runs under a project which matches the filter string and the run_view_type are returned.

        Args:
            ml_repo (str): Name of the project.
            filter_string (str, optional):
                Filter query string, defaults to searching all runs. Identifier required in the LHS of a search expression.
                Signifies an entity to compare against. An identifier has two parts separated by a period: the type of the entity and the name of the entity.
                The type of the entity is metrics, params, attributes, or tags. The entity name can contain alphanumeric characters and special characters.
                You can search using two run attributes : status and artifact_uri. Both attributes have string values.
                When a metric, parameter, or tag name contains a special character like hyphen, space, period, and so on,
                enclose the entity name in double quotes or backticks, params."model-type" or params.`model-type`

            run_view_type (str, optional): one of the following values "ACTIVE_ONLY", "DELETED_ONLY", or "ALL" runs.
            order_by (List[str], optional):
                List of columns to order by (e.g., "metrics.rmse"). Currently supported values
                are metric.key, parameter.key, tag.key, attribute.key. The ``order_by`` column
                can contain an optional ``DESC`` or ``ASC`` value. The default is ``ASC``.
                The default ordering is to sort by ``start_time DESC``.

            job_run_name (str): Name of the job which are associated with the runs to get that runs

            max_results (int): max_results on the total numbers of run yielded through filter

        Examples:
            ```python
            import mlfoundry as mlf

            client = mlf.get_client()
            with client.create_run(ml_repo="my-project", run_name="run-1") as run1:
                run1.log_metrics(metric_dict={"accuracy": 0.74, "loss": 0.6})
                run1.log_params({"model": "LogisticRegression", "lambda": "0.001"})

            with client.create_run(ml_repo="my-project", run_name="run-2") as run2:
                run2.log_metrics(metric_dict={"accuracy": 0.8, "loss": 0.4})
                run2.log_params({"model": "SVM"})

            # Search for the subset of runs with logged accuracy metric greater than 0.75
            filter_string = "metrics.accuracy > 0.75"
            runs = client.search_runs(ml_repo="my-project", filter_string=filter_string)

            # Search for the subset of runs with logged accuracy metric greater than 0.7
            filter_string = "metrics.accuracy > 0.7"
            runs = client.search_runs(ml_repo="my-project", filter_string=filter_string)

            # Search for the subset of runs with logged accuracy metric greater than 0.7 and model="LogisticRegression"
            filter_string = "metrics.accuracy > 0.7 and params.model = 'LogisticRegression'"
            runs = client.search_runs(ml_repo="my-project", filter_string=filter_string)

            # Search for the subset of runs with logged accuracy metric greater than 0.7 and order by accuracy in Descending  order
            filter_string = "metrics.accuracy > 0.7"
            order_by = ["metric.accuracy DESC"]
            runs = client.search_runs(
                ml_repo="my-project", filter_string=filter_string, order_by=order_by
            )

            filter_string = "metrics.accuracy > 0.7"
            runs = client.search_runs(
                ml_repo="transformers", order_by=order_by ,job_run_name='job_run_name', filter_string=filter_string
            )

            order_by = ["metric.accuracy DESC"]
            runs = client.search_runs(
                ml_repo="my-project", filter_string=filter_string, order_by=order_by, max_results=10
            )
            ```

        Returns:
            Genarator[MlFoundryRun, None, None]: MLFoundryRuns matching the search query.
        """
        ml_repo = _resolve_ml_repo_name(ml_repo=ml_repo, project_name=project_name)
        try:
            run_view_type = ViewType.from_string(run_view_type.lower())
        except Exception as e:
            raise MlFoundryException(e) from e

        try:
            ml_repo_obj = self.mlflow_client.get_experiment_by_name(ml_repo)
        except MlflowException as e:
            raise MlFoundryException(e) from e  # user doesnot have READ permission

        if ml_repo_obj is None:
            logger.warning(f"ML Repo with name {ml_repo} does not exist")
            return

        ml_repo_id = ml_repo_obj.experiment_id

        page_token = None
        done = False
        if job_run_name:
            if filter_string == "":
                filter_string = f"tags.TFY_INTERNAL_JOB_RUN_NAME = '{job_run_name}'"
            else:
                filter_string += (
                    f" and tags.TFY_INTERNAL_JOB_RUN_NAME = '{job_run_name}'"
                )
        while not done:
            all_runs = self.mlflow_client.search_runs(
                experiment_ids=[ml_repo_id],
                filter_string=filter_string,
                run_view_type=run_view_type,
                max_results=min(_SEARCH_MAX_RESULTS_DEFAULT, max_results)
                if max_results
                else _SEARCH_MAX_RESULTS_DEFAULT,
                order_by=order_by,
                page_token=page_token,
            )
            page_token = all_runs.token
            for run in all_runs:
                if max_results is not None:
                    max_results -= 1
                yield MlFoundryRun._get_run_from_mlflow_run(run)
            done = page_token is None or max_results == 0

    @staticmethod
    def get_tracking_uri():
        """
        Get the current tracking URI.

        Returns:
            The tracking URI.

        Examples:

            ```python
            import tempfile
            import mlfoundry

            client = mlfoundry.get_client()
            tracking_uri = client.get_tracking_uri()
            print("Current tracking uri: {}".format(tracking_uri))
            ```
        """
        return mlflow.tracking.get_tracking_uri()

    def get_model_version(
        self,
        ml_repo: str,
        name: str,
        version: Union[str, int] = constants.LATEST_ARTIFACT_OR_MODEL_VERSION,
    ) -> Optional[ModelVersion]:
        """
        Get the model version to download contents or load it in memory

        Args:
            ml_repo (str): ML Repo to which model is logged
            name (str): Model Name
            version (str | int): Model Version to fetch (default is the latest version)

        Returns:
            The ModelVersion instance of the model.

        Examples:

            ```python
            import tempfile
            import mlfoundry

            client = mlfoundry.get_client()
            model_version = client.get_model_version(ml_repo="ml-repo-name", name="model-name", version=1)

            # load the model into memory
            clf = model_version.load()

            # download the model to disk
            temp = tempfile.TemporaryDirectory()
            download_info = model_version.download(path=temp.name)
            print(download_info)
            ```
        """
        resolved_version = None
        if version != constants.LATEST_ARTIFACT_OR_MODEL_VERSION:
            resolved_version = _resolve_version(version=version)

        ml_repo_id = self._get_ml_repo(ml_repo=ml_repo)
        return self.mlflow_client.get_model_version(
            experiment_id=int(ml_repo_id),
            model_name=name,
            version=resolved_version,
        )

    def get_model_version_by_fqn(self, fqn: str) -> ModelVersion:
        """
        Get the model version to download contents or load it in memory

        Args:
            fqn (str): Fully qualified name of the model version.

        Returns:
            The ModelVersion instance of the model.

        Examples:

            ```python
            import tempfile
            import mlfoundry

            client = mlfoundry.get_client()
            model_version = client.get_model_version_by_fqn(fqn="model:truefoundry/user/my-classification-project/my-sklearn-model:1")

            # load the model into memory
            clf = model_version.load()

            # download the model to disk
            temp = tempfile.TemporaryDirectory()
            download_info = model_version.download(path=temp.name)
            print(download_info)
            ```
        """
        return ModelVersion.from_fqn(fqn=fqn)

    def get_model(self, fqn: str) -> ModelVersion:
        """
        Get the model version to download contents or load it in memory

        Args:
            fqn (str): Fully qualified name of the model version.

        Returns:
            The ModelVersion instance of the model.

        Examples:

            ```python
            import tempfile
            import mlfoundry

            client = mlfoundry.get_client()
            model_version = client.get_model(fqn="model:truefoundry/user/my-classification-project/my-sklearn-model:1")

            # load the model into memory
            clf = model_version.load()

            # download the model to disk
            temp = tempfile.TemporaryDirectory()
            download_info = model_version.download(path=temp.name)
            print(download_info)
            ```
        """
        warning_message = "get_model will be deprecated in a future release. Please use get_model_version instead"
        warnings.warn(
            message=warning_message, category=DeprecationWarning, stacklevel=2
        )
        return ModelVersion.from_fqn(fqn)

    def list_model_versions(self, ml_repo: str, name: str) -> Iterator[ModelVersion]:
        """
        Get all the version of a model to download contents or load them in memory

        Args:
            ml_repo (str): Repository in which the model is stored.
            name (str): Name of the model whose version is required

        Returns:
            Iterator[ModelVersion]: An iterator that yields non deleted model versions
                of a model under a given ml_repo  sorted reverse by the version number

        Examples:

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            model_versions = client.list_mode_version(ml_repo="my-repo", name="svm")

            for model_version in model_versions:
                print(model_version)
            ```
        """
        ml_repo_id = self._get_ml_repo(ml_repo=ml_repo)
        models = self.mlflow_client.list_models(ml_repo_id=ml_repo_id, name=name)
        if not models or len(models) == 0:
            err_msg = f"Model Does Not Exist for ml_repo={ml_repo}, name={name}"
            raise MlFoundryException(err_msg)
        return self._list_model_versions_by_id(model=models[0])

    def list_model_versions_by_fqn(self, model_fqn: str) -> Iterator[ModelVersion]:
        """
        List versions for a given model

        Args:
            model_fqn: FQN of the Model to list versions for.
                A model_fqn looks like `model:{org}/{user}/{project}/{artifact_name}`
                or `model:{user}/{project}/{artifact_name}`

        Returns:
            Iterator[ModelVersion]: An iterator that yields non deleted model versions
                under the given model_fqn sorted reverse by the version number

        Yields:
            ModelVersion: An instance of `mlfoundry.ModelVersion`

        Examples:

            ```python
            import mlfoundry

            mlfoundry.login(tracking_uri=https://your.truefoundry.site.com")
            client = mlfoundry.get_client()
            model_fqn = "model:org/user/my-project/my-model"
            for mv in client.list_model_versions(model_fqn=model_fqn):
                print(mv.name, mv.version, mv.description)
            ```
        """
        model = self.mlflow_client.get_model_by_fqn(fqn=model_fqn)
        return self._list_model_versions_by_id(model=model)

    def _list_model_versions_by_id(
        self, model_id: str = None, model: Model = None
    ) -> Iterator[ModelVersion]:
        if model and not model_id:
            model_id = model.id
        elif not model and model_id:
            model = self.mlflow_client.get_model_by_id(model_id=model_id)
        else:
            raise MlFoundryException(
                "Exactly one of model_id or model should be passed"
            )

        max_results, page_token, done = 10, None, False
        while not done:
            model_versions = self.mlflow_client.list_model_versions(
                model_id=model_id, max_results=max_results, page_token=page_token
            )
            page_token = model_versions.token
            for model_version in model_versions:
                yield ModelVersion(model_version=model_version, model=model)
            if not model_versions or not page_token:
                done = True

    def get_artifact_version(
        self,
        ml_repo: str,
        artifact_name: str,
        artifact_type: Optional[ArtifactType] = ArtifactType.ARTIFACT,
        version: Union[str, int] = constants.LATEST_ARTIFACT_OR_MODEL_VERSION,
    ) -> Optional[ArtifactVersion]:
        """
        Get the model version to download contents or load it in memory

        Args:
            ml_repo (str): ML Repo to which artifact is logged
            artifact_name (str): Artifact Name
            artifact_type (str): The type of artifact to fetch (acceptable values: "artifact", "model", "plot", "image")
            version (str | int): Artifact Version to fetch (default is the latest version)

        Returns:
            ArtifactVersion : An ArtifactVersion instance of the artifact

        Examples:

            ```python
            import tempfile
            import mlfoundry

            client = mlfoundry.get_client()
            model_version = client.get_artifact_version(ml_repo="ml-repo-name", name="artifact-name", version=1)

            # load the model into memory
            clf = model_version.load()

            # download the model to disk
            temp = tempfile.TemporaryDirectory()
            download_info = model_version.download(path=temp.name)
            print(download_info)
            ```
        """
        resolved_version = None
        if version != constants.LATEST_ARTIFACT_OR_MODEL_VERSION:
            resolved_version = _resolve_version(version=version)

        ml_repo_id = self._get_ml_repo(ml_repo=ml_repo)
        return self.mlflow_client.get_artifact_version(
            experiment_id=int(ml_repo_id),
            artifact_name=artifact_name,
            artifact_type=artifact_type,
            version=resolved_version,
        )

    def get_artifact_version_by_fqn(self, fqn: str) -> ArtifactVersion:
        """
        Get the artifact version to download contents

        Args:
            fqn (str): Fully qualified name of the artifact version.

        Returns:
            ArtifactVersion : An ArtifactVersion instance of the artifact

        Examples:

            ```python
            import tempfile
            import mlfoundry

            client = mlfoundry.get_client()
            artifact_version = client.get_artifact_version_by_fqn(fqn="artifact:truefoundry/user/my-classification-project/sklearn-artifact:1")

            # download the artifact to disk
            temp = tempfile.TemporaryDirectory()
            download_info = artifact_version.download(path=temp.name)
            print(download_info)
            ```
        """
        return ArtifactVersion.from_fqn(fqn=fqn)

    def get_artifact(self, fqn: str) -> ArtifactVersion:
        """
        Get the artifact version to download contents

        Args:
            fqn (str): Fully qualified name of the artifact version.

        Returns:
            ArtifactVersion : An ArtifactVersion instance of the artifact

        Examples:

            ```python
            import tempfile
            import mlfoundry

            client = mlfoundry.get_client()
            artifact_version = client.get_artifact(fqn="artifact:truefoundry/user/my-classification-project/sklearn-artifact:1")

            # download the artifact to disk
            temp = tempfile.TemporaryDirectory()
            download_info = artifact_version.download(path=temp.name)
            print(download_info)
            ```
        """
        warning_message = "get_artifact will be deprecated in a future release. Please use get_artifact_version instead"
        warnings.warn(
            message=warning_message, category=DeprecationWarning, stacklevel=2
        )
        return ArtifactVersion.from_fqn(fqn)

    def list_artifact_versions(
        self,
        ml_repo: str,
        name: str,
        artifact_type: Optional[ArtifactType] = ArtifactType.ARTIFACT,
    ) -> Iterator[ArtifactVersion]:
        """
        Get all the version of a artifact to download contents or load them in memory

        Args:
            ml_repo (str): Repository in which the model is stored.
            name (str): Name of the artifact whose version is required
            artifact_type (ArtifactType): Type of artifact you want for example model, image, etc.

        Returns:
            Iterator[ArtifactVersion]: An iterator that yields non deleted artifact-versions
                of a artifact under a given ml_repo  sorted reverse by the version number

        Examples:

            ```python
            import mlfoundry

            client = mlfoundry.get_client()
            artifact_versions = client.list_artifact_versions(ml_repo="my-repo", name="artifact-name")

            for artifact_version in artifact_versions:
                print(artifact_version)
            ```
        """
        ml_repo_id = self._get_ml_repo(ml_repo=ml_repo)
        artifacts = self.mlflow_client.list_artifacts_(
            ml_repo_id=ml_repo_id, name=name, artifact_type=artifact_type
        )
        if not artifacts or len(artifacts) == 0:
            err_msg = f"Artifact Does Not Exist for ml_repo={ml_repo}, name={name}, type={artifact_type}"
            raise MlFoundryException(err_msg)
        return self._list_artifact_versions_by_id(artifact=artifacts[0])

    def list_artifact_versions_by_fqn(
        self, artifact_fqn: str
    ) -> Iterator[ArtifactVersion]:
        """
        List versions for a given artifact

        Args:
            artifact_fqn: FQN of the Artifact to list versions for.
                An artifact_fqn looks like `{artifact_type}:{org}/{user}/{project}/{artifact_name}`
                or `{artifact_type}:{user}/{project}/{artifact_name}`

                where artifact_type can be on of ("model", "image", "plot")

        Returns:
            Iterator[ArtifactVersion]: An iterator that yields non deleted artifact versions
                under the given artifact_fqn sorted reverse by the version number

        Yields:
            ArtifactVersion: An instance of `mlfoundry.ArtifactVersion`

        Examples:

            ```python
            import mlfoundry

            mlfoundry.login(tracking_uri=https://your.truefoundry.site.com")
            client = mlfoundry.get_client()
            artifact_fqn = "artifact:org/user/my-project/my-artifact"
            for av in client.list_artifact_versions(artifact_fqn=artifact_fqn):
                print(av.name, av.version, av.description)
            ```
        """
        artifact = self.mlflow_client.get_artifact_by_fqn(fqn=artifact_fqn)
        return self._list_artifact_versions_by_id(artifact=artifact)

    def _list_artifact_versions_by_id(
        self, artifact_id: str = None, artifact: Artifact = None
    ) -> Iterator[ArtifactVersion]:
        if artifact and not artifact_id:
            artifact_id = artifact.id
        elif not artifact and artifact_id:
            artifact = self.mlflow_client.get_artifact_by_id(artifact_id=artifact_id)
        else:
            raise MlFoundryException(
                "Exactly one of artifact_id or artifact should be passed"
            )

        max_results, page_token, done = 10, None, False
        while not done:
            artifact_versions = self.mlflow_client.list_artifact_versions(
                artifact_id=artifact_id, max_results=max_results, page_token=page_token
            )
            page_token = artifact_versions.token
            for artifact_version in artifact_versions:
                yield ArtifactVersion(
                    artifact_version=artifact_version, artifact=artifact
                )
            if not artifact_versions or not page_token:
                done = True

    def log_artifact(
        self,
        ml_repo: str,
        name: str,
        artifact_paths: List[
            Union[Tuple[str], Tuple[str, Optional[str]], ArtifactPath]
        ],
        description: Optional[str] = None,
        metadata: Optional[Dict] = None,
    ) -> ArtifactVersion | None:
        """Logs an artifact for the current `ml_repo`.

        An `artifact` is a list of local files and directories. This function packs the mentioned files and directories in `artifact_paths` and uploads them to remote storage linked to the ml_repo

        Args:
            ml_repo (str): Name of the ML Repo to which an artifact is to be logged.
            name (str): Name of the Artifact. If an artifact with this name already exists under the current ml_repo,
                the logged artifact will be added as a new version under that `name`. If no artifact exist with the given
                `name`, the given artifact will be logged as version 1.
            artifact_paths (List[mlfoundry.ArtifactPath], optional): A list of pairs
                of (source path, destination path) to add files and folders
                to the artifact version contents. The first member of the pair should be a file or directory path
                and the second member should be the path inside the artifact contents to upload to.

                ```
                E.g. >>> client.log_artifact(
                     ...     ml_repo="sample-repo",
                     ...     name="xyz",
                     ...     artifact_paths=[
                                mlfoundry.ArtifactPath("foo.txt", "foo/bar/foo.txt"),
                                mlfoundry.ArtifactPath("tokenizer/", "foo/tokenizer/"),
                                mlfoundry.ArtifactPath('bar.text'),
                                ('bar.txt', ),
                                ('foo.txt', 'a/foo.txt')
                             ]
                     ... )
                would result in
                .
                └── foo/
                    ├── bar/
                    │   └── foo.txt
                    └── tokenizer/
                        └── # contents of tokenizer/ directory will be uploaded here
                ```
            description (Optional[str], optional): arbitrary text upto 1024 characters to store as description.
                This field can be updated at any time after logging. Defaults to `None`
            metadata (Optional[Dict[str, Any]], optional): arbitrary json serializable dictionary to store metadata.
                For example, you can use this to store metrics, params, notes.
                This field can be updated at any time after logging. Defaults to `None`

        Returns:
            mlfoundry.ArtifactVersion: an instance of `ArtifactVersion` that can be used to download the files,
            or update attributes like description, metadata.

        Examples:
            ```python
            import os
            import mlfoundry

            with open("artifact.txt", "w") as f:
                f.write("hello-world")

            client = mlfoundry.get_client()
            ml_repo = "sample-repo"

            client.create_ml_repo(ml_repo=ml_repo)
            client.log_artifact(
                ml_repo=ml_repo,
                name="hello-world-file",
                artifact_paths=[mlfoundry.ArtifactPath('artifact.txt', 'a/b/')]
            )
            ```
        """
        if not artifact_paths:
            raise MlFoundryException(
                "artifact_paths cannot be empty, atleast one artifact_path must be passed"
            )

        ml_repo_id = self._get_ml_repo(ml_repo=ml_repo)
        artifact_version = _log_artifact_version(
            run=None,
            mlflow_client=self.mlflow_client,
            ml_repo_id=ml_repo_id,
            name=name,
            artifact_paths=artifact_paths,
            description=description,
            metadata=metadata,
            step=None,
        )
        logger.info(f"Logged artifact successfully with fqn {artifact_version.fqn!r}")
        return artifact_version

    def log_model(
        self,
        *,
        ml_repo: str,
        name: str,
        model: Any,
        framework: Optional[Union[ModelFramework, str]],
        model_save_kwargs: Optional[Dict[str, Any]] = None,
        additional_files: Sequence[Tuple[Union[str, Path], Optional[str]]] = (),
        description: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        model_schema: Optional[Union[Dict[str, Any], ModelSchema]] = None,
        custom_metrics: Optional[List[Union[Dict[str, Any], CustomMetric]]] = None,
    ) -> ModelVersion:
        """
        Serialize and log a versioned model under the current ml_repo. Each logged model generates a new version
        associated with the given `name` and linked to the current run. Multiple versions of the model can be
        logged as separate versions under the same `name`.

        Args:
            ml_repo (str): Name of the ML Repo to which an artifact is to be logged.
            name (str): Name of the model. If a model with this name already exists under the current ml_repo,
                the logged model will be added as a new version under that `name`. If no models exist with the given
                `name`, the given model will be logged as version 1.
            model (Any): model instance of any one of the supported frameworks under `mlfoundry.enums.ModelFramework`.
                Can also be `None` which can be useful to create a reference entry without uploading any model files.
            framework (Union[enums.ModelFramework, str]): Model Framework. Ex:- pytorch, sklearn, tensorflow etc.
                The full list of supported frameworks can be found in `mlfoundry.enums.ModelFramework`.
                Can also be `None` when `model` is `None`.
            model_save_kwargs (Optional[Dict[str, Any]], optional): keyword arguments to pass to model serializer.
                Defaults to `None`
            additional_files (Sequence[Tuple[Union[str, Path], Optional[str]]], optional): A list of pairs
                of (source path, destination path) to add additional files and folders
                to the model version contents. The first member of the pair should be a file or directory path
                and the second member should be the path inside the model versions contents to upload to.
                The model version contents are arranged like follows
                .
                └── model/
                    └── # model files are serialized here
                └── # any additional files and folders can be added here.

                You can also add additional files to model/ subdirectory by specifying the destination path as model/
                ```
                E.g. >>> client.log_model(
                     ...     ml_repo="sample-repo", name="xyz", model=clf, framework="sklearn",
                     ...     additional_files=[("foo.txt", "foo/bar/foo.txt"), ("tokenizer/", "foo/tokenizer/")]
                     ... )
                would result in
                .
                ├── model/
                │   └── # model files are serialized here e.g. model.joblib
                └── foo/
                    ├── bar/
                    │   └── foo.txt
                    └── tokenizer/
                        └── # contents of tokenizer/ directory will be uploaded here
                ```
            description (Optional[str], optional): arbitrary text upto 1024 characters to store as description.
                This field can be updated at any time after logging. Defaults to `None`
            metadata (Optional[Dict[str, Any]], optional): arbitrary json serializable dictionary to store metadata.
                For example, you can use this to store metrics, params, notes.
                This field can be updated at any time after logging. Defaults to `None`
            model_schema (Optional[Union[Dict[str, Any], ModelSchema]], optional): instance of `mlfoundry.ModelSchema`.
                This schema needs to be consistent with older versions of the model under the given `name` i.e.
                a feature's value type and model's prediction type cannot be changed in the schema of new version.
                Features can be removed or added between versions.
                ```
                E.g. if there exists a v1 with
                schema = {"features": {"name": "feat1": "int"}, "prediction": "categorical"}, then

                schema = {"features": {"name": "feat1": "string"}, "prediction": "categorical"} or
                schema = {"features": {"name": "feat1": "int"}, "prediction": "numerical"}
                are invalid because they change the types of existing features and prediction

                while
                schema = {"features": {"name": "feat1": "int", "feat2": "string"}, "prediction": "categorical"} or
                schema = {"features": {"feat2": "string"}, "prediction": "categorical"}
                are valid

                This field can be updated at any time after logging. Defaults to `None`
                ```
            custom_metrics: (Optional[Union[List[Dict[str, Any]], CustomMetric]], optional): list of instances of
                `mlfoundry.CustomMetric`
                The custom metrics must be added according to the prediction type of schema.
                custom_metrics = [{
                    "name": "mean_square_error",
                    "type": "metric",
                    "value_type": "float"
                }]

        Returns:
            mlfoundry.ModelVersion: an instance of `ModelVersion` that can be used to download the files,
                load the model, or update attributes like description, metadata, schema.

        Examples:

            ### sklearn
            ```python
            import mlfoundry
            import numpy as np
            from sklearn.pipeline import make_pipeline
            from sklearn.preprocessing import StandardScaler
            from sklearn.svm import SVC

            client = mlfoundry.get_client()
            run = client.create_run(
                ml_repo="my-classification-project"
            )
            X = np.array([[-1, -1], [-2, -1], [1, 1], [2, 1]])
            y = np.array([1, 1, 2, 2])
            clf = make_pipeline(StandardScaler(), SVC(gamma='auto'))
            clf.fit(X, y)

            model_version = run.log_model(
                name="my-sklearn-model",
                model=clf,
                framework="sklearn"
            )
            print(model_version.fqn)
            ```
        """
        ml_repo_id = self._get_ml_repo(ml_repo=ml_repo)

        model_version = _log_model_version(
            run=None,
            mlflow_client=self.mlflow_client,
            ml_repo_id=ml_repo_id,
            name=name,
            model=model,
            framework=framework,
            model_save_kwargs=model_save_kwargs,
            additional_files=additional_files,
            description=description,
            metadata=metadata,
            model_schema=model_schema,
            custom_metrics=custom_metrics,
            step=None,
        )
        logger.info(f"Logged model successfully with fqn {model_version.fqn!r}")
        return model_version



from mlflow.utils.rest_utils import MlflowHostCreds, http_request_safe

from mlfoundry.run_utils import append_servicefoundry_path_to_tracking_ui
from mlfoundry.tracking.entities import Token


class ServicefoundryService:
    def __init__(self, tracking_uri: str, token: str = None):
        self.host_creds = MlflowHostCreds(
            host=append_servicefoundry_path_to_tracking_ui(tracking_uri), token=token
        )

    def get_token_from_api_key(self, api_key: str) -> Token:
        response = http_request_safe(
            host_creds=self.host_creds,
            endpoint="/v1/oauth/api-key/token",
            method="get",
            params={"apiKey": api_key},
            timeout=3,
            max_retries=0,
        )
        response = response.json()
        return Token.parse_obj(response)

    def get_integration_from_id(self, integration_id: str):
        integration_id = integration_id or ""
        response = http_request_safe(
            host_creds=self.host_creds,
            endpoint=f"/v1/integrations/{integration_id}",
            method="get",
            timeout=3,
            max_retries=0,
        )
        return response.json()



import typing
from collections import defaultdict
from datetime import datetime
from functools import lru_cache

from mlflow.utils.rest_utils import MlflowHostCreds, http_request_safe

from mlfoundry.logger import logger
from mlfoundry.monitoring.entities import (
    ActualPacket,
    BasePacket,
    DatasetData,
    ModelVersion,
    PredictionPacket,
)
from mlfoundry.monitoring.store.repositories.dto import (
    ActualData,
    BatchInsertRequest,
    BatchUpdateActualRequest,
    ClassPrediction,
    Data,
    GetDatasetResponse,
    MlModelPrediction,
)


# TODO @nikp1172 add feature of autmatically serializing datetime object in \
#  mlflow.utils.rest_utils.http_request
class RestMonitoringStore:
    def __init__(self, get_host_creds: typing.Callable[[], MlflowHostCreds]):
        self._get_host_creds: typing.Callable[[], MlflowHostCreds] = get_host_creds

    def _insert_prediction(self, prediction_packets: typing.List[PredictionPacket]):
        if not prediction_packets:
            return
        # group packets by model version id
        grouped_packets = defaultdict(list)

        for packet in prediction_packets:
            grouped_packets[packet.model_version_id].append(packet)

        for model_version_id in grouped_packets.keys():
            items = [
                Data(
                    data_id=packet.prediction.data_id,
                    features=packet.prediction.features,
                    prediction=MlModelPrediction(
                        value=packet.prediction.prediction_data.value,
                        probabilities=[
                            ClassPrediction(label=feature_name, score=score)
                            for feature_name, score in packet.prediction.prediction_data.probabilities.items()
                        ],
                        shap_values=packet.prediction.prediction_data.shap_values,
                        occurred_at=packet.prediction.occurred_at,
                    ),
                    actual=packet.prediction.actual_value,
                    raw_data=packet.prediction.raw_data,
                )
                for packet in grouped_packets[model_version_id]
            ]

            batch_insert_request = BatchInsertRequest(
                model_version_id=model_version_id,
                items=items,
            )

            response = http_request_safe(
                host_creds=self._get_host_creds(),
                endpoint="/v1/data/batch-insert",
                method="post",
                json=batch_insert_request.to_json_dict(),
            )

    def _update_actuals(self, actual_packets: typing.List[ActualPacket]):
        if not actual_packets:
            return

        grouped_packets = defaultdict(list)

        for packet in actual_packets:
            grouped_packets[packet.model_version_id].append(packet)

        for model_version_id in grouped_packets.keys():
            batch_update_request = BatchUpdateActualRequest(
                model_version_id=model_version_id,
                items=[
                    ActualData(
                        data_id=packet.actual.data_id, actual=packet.actual.value
                    )
                    for packet in grouped_packets[model_version_id]
                ],
            )
            response = http_request_safe(
                host_creds=self._get_host_creds(),
                endpoint="/v1/data/batch-update-actual",
                method="post",
                json=batch_update_request.dict(),
            )

    def batch_log_inference(self, inference_packets: typing.List[BasePacket]):
        grouped_packets = defaultdict(list)
        for packet in inference_packets:
            grouped_packets[type(packet)].append(packet)

        # log predictions before actuals to avoid actuals getting logged before predictions
        if PredictionPacket in grouped_packets.keys():
            self._insert_prediction(grouped_packets[PredictionPacket])
        if ActualPacket in grouped_packets.keys():
            self._update_actuals(grouped_packets[ActualPacket])

    @lru_cache(maxsize=32)
    def enable_monitoring_for_version(self, model_version_fqn: str):
        response = http_request_safe(
            host_creds=self._get_host_creds(),
            endpoint=f"/v1/model/enable",
            method="post",
            json={"model_version_fqn": model_version_fqn},
        )
        response_json = response.json()
        return ModelVersion(**response_json)

    def _get_inference_data(
        self,
        model_id: str,
        start_time: datetime,
        end_time: datetime,
        actual_value_required: bool,
        offset: int,
        limit: int,
    ) -> GetDatasetResponse:
        response = http_request_safe(
            host_creds=self._get_host_creds(),
            endpoint="/v1/data/get-dataset",
            method="post",
            json={
                "model_id": model_id,
                "start_time": start_time.isoformat(),
                "end_time": end_time.isoformat(),
                "actual_value_required": actual_value_required,
                "offset": offset,
                "limit": limit,
            },
        )
        response_json = response.json()
        return GetDatasetResponse(**response_json)

    def get_inference_dataset(
        self,
        model_id: str,
        start_time: datetime,
        end_time: datetime,
        actual_value_required: bool,
    ) -> typing.List[DatasetData]:

        total_results = None
        data = []

        filters = []
        if actual_value_required:
            filters.append({"type": "neq", "dimension": "actual.value", "value": None})

        while total_results is None or len(data) < total_results:
            list_dataset_data_object = self._get_inference_data(
                model_id=model_id,
                start_time=start_time,
                end_time=end_time,
                actual_value_required=actual_value_required,
                offset=len(data),
                limit=5000,
            )
            if not total_results:
                total_results = list_dataset_data_object.total_rows
                logger.info(f"Total Results: {total_results} for model_id: {model_id}")
            data.extend(list_dataset_data_object.data)
            if not len(list_dataset_data_object.data):
                break
            logger.info(
                f"Fetched {len(list_dataset_data_object.data)} rows for {model_id}."
            )
        return data



import enum
import platform
import time
import typing
import uuid

import amplitude_tracker

from mlfoundry.logger import logger

try:
    from git.config import GitConfigParser
except Exception as ex:
    GitConfigParser = None


# API key for mlfoundry project
AMPLITUDE_API_KEY = "e4925dfc6d2831828b3e55ae83e19fd2"
NO_USER = "no_user"
EventPropertiesType = typing.Dict[str, str]
CHAR_LIMIT = 1024
__all__ = ["Event", "init", "track"]


@enum.unique
class Event(enum.Enum):
    GET_CLIENT = "GET_CLIENT"
    GET_MONITORING_CLIENT = "GET_MONITORING_CLIENT"
    EXCEPTION = "EXCEPTION"
    #### Top level APIs
    GET_ALL_PROJECTS = "GET_ALL_PROJECTS"
    CREATE_RUN = "CREATE_RUN"
    GET_RUN = "GET_RUN"
    GET_ALL_RUNS = "GET_ALL_RUNS"
    ####
    #### Run Level APIs
    LOG_MODEL = "LOG_MODEL"
    LOG_DATASET = "LOG_DATASET"
    LOG_METRICS = "LOG_METRICS"
    LOG_PARAMS = "LOG_PARAMS"
    GET_DATASET = "GET_DATASET"
    GET_METRICS = "GET_METRICS"
    GET_PARAMS = "GET_PARAMS"
    GET_MODEL = "GET_MODEL"
    GET_TAGS = "GET_TAGS"
    SET_TAGS = "SET_TAGS"
    ####


class _AmplitudeClient:
    def __init__(self, user_id: str, api_key: str = AMPLITUDE_API_KEY) -> None:
        amplitude_tracker.write_key = api_key
        self.device_id = str(uuid.getnode())
        self.platform = platform.system()
        self.session_id = round(time.time() * 1000)
        self.user_id = user_id
        try:
            self.user_properties = {
                "python_version": platform.python_version(),
                # Ideally this should not be in user_properties
                # the amplitude_tracker overwrites the top level platform key
                # https://github.com/RandomCoffee/amplitude-python/blob/b833ffeeefa18eb69d2f2378ec6e13eb2587cdfa/amplitude_tracker/client.py#L93
                "platform": self.platform,
                "platform_info": platform.platform(terse=True),
            }
        except Exception as ex:
            logger.warning(f"failed to get user properties {ex}")
            self.user_properties = {}

    def track(
        self,
        event: Event,
        event_properties: typing.Optional[EventPropertiesType] = None,
    ):
        event_properties = event_properties if event_properties is not None else {}
        try:
            for key in event_properties:
                event_properties[key] = str(event_properties[key])[:CHAR_LIMIT]
            amplitude_tracker.track(
                user_id=self.user_id,
                device_id=self.device_id,
                session_id=self.session_id,
                event_type=event.value,
                event_properties=event_properties,
                user_properties=self.user_properties,
            )
        except Exception as ex:
            logger.warning(f"failed to track event, {ex}")


class _FakeClient(_AmplitudeClient):
    def __init__(self, *args, **kwargs):
        ...

    def track(self, *args, **kwargs):
        ...


_amplitude_client: typing.Optional[_AmplitudeClient] = None


def init(user_id: str, disable_analytics: bool = False):
    global _amplitude_client
    if _amplitude_client is not None:
        logger.debug("amplitude already initialized")
        return
    if disable_analytics:
        logger.info("analytics is disabled")
        _amplitude_client = _FakeClient()
    else:
        _amplitude_client = _AmplitudeClient(user_id=user_id)


def track(
    event: Event,
    event_properties: typing.Optional[EventPropertiesType] = None,
):
    if _amplitude_client is None:
        logger.warning("Amplitude client is not initialised, please call `init` first")
        return
    _amplitude_client.track(event=event, event_properties=event_properties)



import argparse
import enum
import json
import os
import tempfile
from typing import Callable, Dict, List, Mapping, Optional, Sequence, Union
from weakref import ReferenceType

from mlflow.entities import RunStatus
from pytorch_lightning.callbacks.model_checkpoint import ModelCheckpoint
from pytorch_lightning.loggers.base import LightningLoggerBase, rank_zero_experiment
from pytorch_lightning.utilities.distributed import rank_zero_only
from pytorch_lightning.utilities.logger import (
    _add_prefix,
    _convert_params,
    _sanitize_callable_params,
)

import mlfoundry
from mlfoundry.enums import EnumMissingMixin
from mlfoundry.logger import logger
from mlfoundry.mlfoundry_api import _resolve_ml_repo_name
from mlfoundry.mlfoundry_run import MlFoundryRun

LIGHTNING_ARTIFACTS_PATH = "lightning/"
LIGHTNING_MODELS_PATH = os.path.join(LIGHTNING_ARTIFACTS_PATH, "models")


class LogModelStrategy(EnumMissingMixin, enum.Enum):
    # To keep it consistent with transformers
    NO = "no"
    BEST_ONLY = "best"
    BEST_PLUS_LATEST = "best_plus_latest"
    CHECKPOINTS_ON_TRAIN_END = "checkpoints_on_train_end"
    # TODO: Add ALL strategy, and how to handle that?


def _get_artifact_path_for_checkpoint(checkpoint_path):
    return os.path.join(LIGHTNING_MODELS_PATH, os.path.basename(checkpoint_path))


class MlFoundryLightningLogger(LightningLoggerBase):
    """
    Pytorch Lightning Logger for tracking training run on MLFoundry

    Examples:

        ```
        import mlfoundry
        mlfoundry.login()
        # or set API key via `TFY_API_KEY` env variable

        from pytorch_lightning import Trainer
        from pytorch_lightning.callbacks import ModelCheckpoint
        from mlfoundry.integrations.lightning import MlFoundryLightningLogger, LogModelStrategy

        mlf_logger = MlFoundryLightningLogger(
            project_name="lightning",
            run_name="my-lightning-run",
            flatten_params=True,
            log_model_strategy=LogModelStrategy.BEST_PLUS_LATEST,
        )

        checkpoint_callback = ModelCheckpoint(
            monitor='val_accuracy',
            mode='max',
            save_last=True,
            save_top_k=-1
            )
        trainer = Trainer(
            logger=mlf_logger,
            callbacks=[checkpoint_callback], # our model checkpoint callback
            ..
            )
        trainer.fit(model, training_loader, validation_loader)
        ```

        Logger can also be created from an existing run

        ```
        import mlfoundry
        mlfoundry.login()
        # or set API key via `TFY_API_KEY` env variable

        from pytorch_lightning import Trainer
        from pytorch_lightning.callbacks import ModelCheckpoint
        from mlfoundry.integrations.lightning import MlFoundryLightningLogger, LogModelStrategy

        client = mlfoundry.get_client()
        run = client.create_run(project_name="lightning")

        mlf_logger = MlFoundryLightningLogger.from_run(
            run=run,
            auto_end_run=False,
            flatten_params=True,
            log_model_strategy=LogModelStrategy.BEST_PLUS_LATEST,
        )

        checkpoint_callback = ModelCheckpoint(
            monitor='val_accuracy',
            mode='max',
            save_last=True,
            save_top_k=-1
            )
        trainer = Trainer(
            logger=mlf_logger,
            callbacks=[checkpoint_callback], # our model checkpoint callback
            ..
            )
        trainer.fit(model, training_loader, validation_loader)

        run.end()
        ```
    """

    def __init__(
        self,
        ml_repo: Optional[str] = None,
        run_name: Optional[str] = None,
        log_model_strategy: Union[str, LogModelStrategy] = LogModelStrategy.NO,
        flatten_params: Optional[bool] = False,
        prefix: str = "",
        agg_key_funcs: Optional[
            Mapping[str, Callable[[Sequence[float]], float]]
        ] = None,
        agg_default_func: Optional[Callable[[Sequence[float]], float]] = None,
        project_name: Optional[str] = None,
        **kwargs,
    ):
        """
        Args:
            ml_repo (str): name of the ML repo to create the run under
            run_name (Optional[str], optional): name of the run. When not provided a run name is automatically generated
            log_model_strategy (LogModelStrategy, optional): The strategy to use for logging models
                - LogModelStrategy.NO (default): Do not log any models
                - LogModelStrategy.BEST_ONLY: Log only the best model checkpoint.
                - LogModelStrategy.BEST_PLUS_LATEST: Log both the latest checkpoint and the best checkpoint (if
                    available) and different from the latest checkpoint. Note that `save_last` argument must be set
                    in the `ModelCheckpoint` callback, else latest checkpoints wont be saved/logged.
                - LogModelStrategy.CHECKPOINTS_ON_TRAIN_END: Log all available checkpoints that are available at the
                    end of training. Note that `save_top_k` must be set in the `ModelCheckpoint` callback,
                    else models from the previous steps wont be saved/logged

                In addition to checkpoints a metadata.json file is logged which stores references to the best checkpoint
                and the latest checkpoint and their corresponding scores

                The files will be logged under `mlf/lightning/models/` under the run artifacts.
            flatten_params (bool, optional): Flatten hierarchical dict
                By default, this is `False`

                For e.g.
                when set to `True`:
                `{'a': {'b': 'c'}}` will be logged as -> `{'a/b': 'c'}`

                when set to `False`:
                Individual keys will be logged ie) `'a':{'b':'c'}`
            prefix (str, optional): This prefix will be added in front of all metric_name logged.

        """
        super().__init__(agg_key_funcs=agg_key_funcs, agg_default_func=agg_default_func)

        log_model_strategy = LogModelStrategy(log_model_strategy)

        self._ml_repo = _resolve_ml_repo_name(
            ml_repo=ml_repo, project_name=project_name
        )
        self._run_name = run_name
        self._log_model_strategy = log_model_strategy
        self._flatten_params = flatten_params
        self._auto_end_run = True
        self._prefix = prefix

        self._run: Optional[MlFoundryRun] = None
        self._checkpoint_callback = None
        self._available_model_paths: List[str] = []

    @property
    @rank_zero_experiment  # Returns the real experiment on rank 0 and otherwise the DummyExperiment
    def run(self) -> MlFoundryRun:
        if self._run:
            return self._run

        if not self._run:
            self._run = mlfoundry.get_client().create_run(
                ml_repo=self._ml_repo,
                run_name=self._run_name,
            )
            self._run_name = self._run.run_name
            self._auto_end_run = True

        return self._run

    @property
    def name(self) -> str:
        return self.run.ml_repo

    @property
    def version(self) -> str:
        return self.run.run_name

    @classmethod
    def from_run(
        cls,
        run: MlFoundryRun,
        log_model_strategy: Union[str, LogModelStrategy] = LogModelStrategy.NO,
        flatten_params: Optional[bool] = False,
        auto_end_run: Optional[bool] = False,
        **kwargs,
    ):
        """
        Create a MLFoundry Lightning Logger from an existing MLFoundry run instance.

        Args:
            run (MlFoundryRun): `MlFoundry` run instance
            log_model_strategy (LogModelStrategy, optional): The strategy to use for logging models
                - LogModelStrategy.NO (default): Do not log any models
                - LogModelStrategy.BEST_ONLY: Log only the best model checkpoint.
                - LogModelStrategy.BEST_PLUS_LATEST: Log both the latest checkpoint and the best checkpoint (if
                    available) and different from the latest checkpoint. Note that `save_last` argument must be set
                    in the `ModelCheckpoint` callback, else latest checkpoints wont be saved/logged.
                - LogModelStrategy.CHECKPOINTS_ON_TRAIN_END: Log all available checkpoints that are available at the
                    end of training. Note that `save_top_k` must be set in the `ModelCheckpoint` callback,
                    else models from the previous steps wont be saved/logged

                In addition to checkpoints a metadata.json file is logged which stores references to the best checkpoint
                and the latest checkpoint and their corresponding scores

                The files will be logged under `mlf/lightning/models/` under the run artifacts.
            flatten_params (bool, optional): Flatten hierarchical dict
                By default, this is `False`

                For e.g.
                when set to `True`:
                `{'a': {'b': 'c'}}` will be logged as -> `{'a/b': 'c'}`

                when set to `False`:
                Individual keys will be logged ie) `'a':{'b':'c'}`
            auto_end_run (bool, optional): if to end the run when training finishes. By default, this is `False`
        """
        instance = cls(
            project_name=run.ml_repo,
            run_name=run.run_name,
            log_model_strategy=log_model_strategy,
            flatten_params=flatten_params,
            **kwargs,
        )
        instance._run = run
        instance._auto_end_run = auto_end_run

        return instance

    @rank_zero_only  # runs only on global rank 0. To prevent multiple logging if there is distributed training
    def log_hyperparams(self, params: argparse.Namespace, *args, **kwargs) -> None:
        """Record hyperparameters.

        Args:
            params (argparse.Namespace): containing the hyperparameters
            args: Optional positional arguments, depends on the specific logger being used
            kwargs: Optional keyword arguments, depends on the specific logger being used
        """
        params = _convert_params(params)  # convert to dict
        params = _sanitize_callable_params(params)  # convert callable params

        self.run.log_params(param_dict=params, flatten_params=self._flatten_params)

    @rank_zero_only
    def log_metrics(
        self, metrics: Dict[str, float], step: Optional[int] = None
    ) -> None:
        """
        Records metrics.
        This method logs metrics as soon as it received them

        Args:
            metrics: Dictionary with metric names as keys and measured quantities as values
            step: Step number at which the metrics should be recorded, defaults to 0
        """
        step = step or 0
        metrics = _add_prefix(metrics, self._prefix, "/")
        self.run.log_metrics(metric_dict=metrics, step=step)

    def _log_model_on_train_end(
        self, checkpoint_callback: "ReferenceType[ModelCheckpoint]", temp_dir
    ):

        latest_checkpoint_path = checkpoint_callback.last_model_path
        best_checkpoint_path = checkpoint_callback.best_model_path
        metadata = {"checkpoints": {}}

        checkpoints_to_log: List[str] = []
        if self._log_model_strategy == LogModelStrategy.BEST_ONLY:
            latest_checkpoint_path = None
            if best_checkpoint_path is not None and os.path.exists(
                best_checkpoint_path
            ):
                checkpoints_to_log.append(best_checkpoint_path)
            else:
                logger.warning(
                    f"`log_model_strategy` was set to BEST_ONLY but CheckpointCallback did not save any best "
                    f"checkpoint. Cannot automatically log model"
                )

        elif self._log_model_strategy == LogModelStrategy.CHECKPOINTS_ON_TRAIN_END:
            for model_path in self._available_model_paths:
                if os.path.exists(model_path):
                    checkpoints_to_log.append(model_path)

        elif self._log_model_strategy == LogModelStrategy.BEST_PLUS_LATEST:
            if latest_checkpoint_path and os.path.exists(latest_checkpoint_path):
                checkpoints_to_log.append(latest_checkpoint_path)
            if (
                best_checkpoint_path
                and best_checkpoint_path != latest_checkpoint_path
                and os.path.exists(best_checkpoint_path)
            ):
                checkpoints_to_log.append(best_checkpoint_path)
        else:
            raise NotImplementedError("Unreachable!")

        for checkpoint_path in checkpoints_to_log:
            checkpoint_name = os.path.basename(checkpoint_path)
            artifact_path = os.path.join(LIGHTNING_MODELS_PATH, checkpoint_name)
            logger.debug(
                f"Logging checkpoint at {checkpoint_path!r} to {artifact_path!r}"
            )
            self._run.log_artifact_deprecated(
                local_path=checkpoint_path, artifact_path=artifact_path
            )

        if best_checkpoint_path:
            best_model_artifact_path = _get_artifact_path_for_checkpoint(
                os.path.basename(best_checkpoint_path)
            )
            metadata["checkpoints"]["best"] = best_model_artifact_path

        if latest_checkpoint_path:
            latest_model_artifact_path = _get_artifact_path_for_checkpoint(
                os.path.basename(latest_checkpoint_path)
            )
            metadata["checkpoints"]["latest"] = latest_model_artifact_path

        metadata_file_name = "metadata.json"
        metadata_path = os.path.join(temp_dir, metadata_file_name)
        with open(metadata_path, "w") as f:
            json.dump(metadata, f)
        self._run.log_artifact_deprecated(
            local_path=metadata_path,
            artifact_path=LIGHTNING_MODELS_PATH,
        )

    def after_save_checkpoint(
        self, checkpoint_callback: "ReferenceType[ModelCheckpoint]"
    ) -> None:
        """Event called after model checkpoint callback saves a new checkpoint.

        Args:
            checkpoint_callback: the model checkpoint callback instance
        """
        if self._log_model_strategy != LogModelStrategy.NO:
            self._checkpoint_callback = checkpoint_callback

            checkpoints = [
                checkpoint_callback.last_model_path,
                checkpoint_callback.best_model_path,
            ] + list(checkpoint_callback.best_k_models.keys())

            for model_path in checkpoints:
                if model_path not in self._available_model_paths:
                    self._available_model_paths.append(model_path)

    def finalize(self, status: str) -> None:
        """Event called at the end of training. Do any processing that is necessary to finalize an run.

        Args:
            status: Status that the run finished with (e.g. success, failed, aborted)
        """
        if self._checkpoint_callback:
            with tempfile.TemporaryDirectory() as temp_dir:
                self._log_model_on_train_end(self._checkpoint_callback, temp_dir)

        if self._auto_end_run and self.run:
            if status == "success":
                self.run.end()
            else:
                self.run.end(status=RunStatus.FAILED)

    def __del__(self):
        if self._auto_end_run and self.run:
            self.run.end()



import json
import logging
import os
import posixpath
import shutil
import uuid
from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, Optional, Sequence, Tuple, Union

import mlflow
from mlflow.tracking import MlflowClient

from mlfoundry.exceptions import MlFoundryException
from mlfoundry.log_types.artifacts.constants import DESCRIPTION_MAX_LENGTH

logger = logging.getLogger("mlfoundry")


def _copy_tree(src_path, dest_path, symlinks=False, ignore_dangling_symlinks=False):
    os.makedirs(dest_path, exist_ok=True)
    for item in os.listdir(src_path):
        src = os.path.join(src_path, item)
        dest = os.path.join(dest_path, item)
        if os.path.isdir(src):
            _copy_tree(
                src,
                dest,
                symlinks=symlinks,
                ignore_dangling_symlinks=ignore_dangling_symlinks,
            )
        else:
            shutil.copy2(src, dest, follow_symlinks=True)


def _get_mlflow_client() -> MlflowClient:
    tracking_uri = mlflow.get_tracking_uri()
    if not tracking_uri:
        # TODO (chiragjn): We should link docs here
        raise MlFoundryException(
            "mlfoundry client needs to be configured before ModelVersion can be initialised. Please use"
            "mlfoundry.get_client() to configure and initialise mlfoundry"
        )
    return MlflowClient()


def is_destination_path_dirlike(dest_path) -> bool:
    if not dest_path:
        return True

    if dest_path.endswith(os.sep) or dest_path.endswith(posixpath.sep):
        return True

    if os.path.exists(dest_path) and os.path.isdir(dest_path):
        return True

    return False


def _copy_additional_files(
    root_dir: str,
    files_dir: str,  # relative to root dir e.g. "files/"
    model_dir: Optional[str],  # relative to files_dir e.g "model/"
    additional_files: Sequence[Tuple[Union[str, Path], Optional[str]]],
):
    """

    File copying examples:
        # non ambiguous
        # a.txt -> /tmp/                            result /tmp/a.txt
        # a.txt -> /tmp/a/                          result /tmp/a/a.txt
        # a.txt -> /tmp/a/b/c/d.txt                 result /tmp/a/b/c/d.txt
        # .gitignore -> /tmp/.gitignore             result /tmp/.gitignore

        # ambiguous but destination directory exists
        # a.txt -> /tmp                             result /tmp/a.txt
        # a.txt -> /tmp/a (and /tmp/a/ exists)      result /tmp/a/a.txt

        # ambiguous - when the destination can't be reliably distinguished as a directory
        # a -> /tmp/a                                result /tmp/a
        # a -> /tmp/b                                result /tmp/b
        # a -> /tmp/a.txt                            result /tmp/a.txt
        # .gitignore -> /tmp/.gitinclude             result /tmp/.gitinclude
        # a.txt -> /tmp/a                            result /tmp/a
    """
    for src_path, dest_path in additional_files:
        src_path = str(src_path)
        if not os.path.exists(src_path):
            raise MlFoundryException(
                f"Source path {src_path!r} in `additional_files` does not exist."
            )
        dest_path = dest_path or ""
        normalized_path = os.path.normpath(dest_path)
        if dest_path.endswith(os.sep) or dest_path.endswith(posixpath.sep):
            normalized_path += os.sep
        dest_path = normalized_path.lstrip(os.sep)

        if model_dir and dest_path.startswith(model_dir):
            logger.warning(
                f"Destination path {dest_path!r} in `additional_files` conflicts with "
                f"reserved path {model_dir!r}/ which is being used to serialize the model. "
                f"This might cause errors"
            )

        files_abs_dir = os.path.join(root_dir, files_dir)
        dest_abs_path = os.path.join(files_abs_dir, dest_path)

        if os.path.isfile(src_path):
            _src = src_path
            if is_destination_path_dirlike(dest_abs_path):
                os.makedirs(dest_abs_path, exist_ok=True)
                _dst = os.path.relpath(
                    os.path.join(dest_abs_path, os.path.basename(_src)), files_abs_dir
                )
            else:
                os.makedirs(os.path.dirname(dest_abs_path), exist_ok=True)
                _dst = os.path.relpath(dest_abs_path, files_abs_dir)
            logger.info(f"Adding file {_src} as /{_dst}")
            shutil.copy2(src_path, dest_abs_path, follow_symlinks=True)
        elif os.path.isdir(src_path):
            os.makedirs(dest_abs_path, exist_ok=True)
            _src = src_path.rstrip("/")
            _dst = os.path.relpath(dest_abs_path, files_abs_dir).rstrip("/")
            logger.info(f"Adding contents of {_src}/ to /{_dst}/")
            _copy_tree(
                src_path=src_path,
                dest_path=dest_abs_path,
                symlinks=True,
                ignore_dangling_symlinks=False,
            )


def _validate_description(description: Optional[str]):
    if description is not None:
        if not isinstance(description, str):
            raise MlFoundryException(
                f"`description` must be either `None` or type `str`"
            )
        if len(description) > DESCRIPTION_MAX_LENGTH:
            raise MlFoundryException(
                f"`description` cannot be longer than {DESCRIPTION_MAX_LENGTH} characters"
            )


def _validate_artifact_metadata(metadata: Dict[str, Any]):
    if not isinstance(metadata, dict):
        raise MlFoundryException("`metadata` must be json serializable dict")
    try:
        json.dumps(metadata)
    except ValueError as ve:
        raise MlFoundryException("`metadata` must be json serializable dict") from ve


@lru_cache(maxsize=32)
def get_model_version_id_from_fqn(fqn) -> uuid.UUID:
    client = _get_mlflow_client()
    model_version = client.get_model_version_by_fqn(fqn=fqn)
    return model_version.id



from __future__ import annotations

import abc
import atexit
import os
import threading
import weakref
from functools import partial
from typing import TYPE_CHECKING, Callable, Dict, Optional

import mlflow
from mlflow.store.artifact.artifact_repository_registry import (
    _artifact_repository_registry,
)
from mlflow.store.tracking.rest_store import RestStore
from mlflow.tracking._tracking_service.utils import _tracking_store_registry
from mlflow.utils.rest_utils import MlflowHostCreds

from mlfoundry import env_vars
from mlfoundry.artifact.truefoundry_artifact_repo import TruefoundryArtifactRepository
from mlfoundry.exceptions import MlFoundryException
from mlfoundry.logger import logger
from mlfoundry.login import CredentialsFileContent, CredentialsFileManager
from mlfoundry.run_utils import (
    append_monitoring_path_to_rest_tracking_uri,
    append_path_to_rest_tracking_uri,
    resolve_tracking_uri,
)
from mlfoundry.tracking.entities import Token, UserInfo
from mlfoundry.tracking.servicefoundry_service import ServicefoundryService
from mlfoundry.tracking.truefoundry_rest_store import (
    TruefoundryRestStore,
    get_rest_store,
)

if TYPE_CHECKING:
    from mlfoundry.mlfoundry_run import MlFoundryRun

TOKEN_REFRESH_LOCK = threading.RLock()


class CredentialProvider(abc.ABC):
    @property
    @abc.abstractmethod
    def token(self) -> Token:
        ...

    @staticmethod
    @abc.abstractmethod
    def can_provide() -> bool:
        ...

    @property
    @abc.abstractmethod
    def tracking_uri(self) -> str:
        ...


class EnvCredentialProvider(CredentialProvider):
    def __init__(self):
        logger.debug("Using env var credential provider")
        self._tracking_uri = resolve_tracking_uri(tracking_uri=None)
        self._auth_service = get_rest_store(
            tracking_uri=self._tracking_uri
        ).get_auth_service()
        api_key = os.getenv(env_vars.API_KEY_GLOBAL)
        if not api_key:
            raise MlFoundryException(
                f"Value of {env_vars.API_KEY_GLOBAL} env var should be non-empty string"
            )
        servicefoundry_service = ServicefoundryService(tracking_uri=self._tracking_uri)
        self._token: Token = servicefoundry_service.get_token_from_api_key(api_key)

    @staticmethod
    def can_provide() -> bool:
        return env_vars.API_KEY_GLOBAL in os.environ

    @property
    def token(self) -> Token:
        with TOKEN_REFRESH_LOCK:
            if self._token.is_going_to_be_expired():
                logger.info("Refreshing access token")
                self._token = self._auth_service.refresh_token(self._token)
            return self._token

    @property
    def tracking_uri(self) -> str:
        return self._tracking_uri


class FileCredentialProvider(CredentialProvider):
    def __init__(self):
        logger.debug("Using file credential provider")
        self._cred_file = CredentialsFileManager()

        with self._cred_file:
            self._last_cred_file_content = self._cred_file.read()
            self._tracking_uri = self._last_cred_file_content.host
            self._token = self._last_cred_file_content.to_token()
            self._auth_service = get_rest_store(
                tracking_uri=self._tracking_uri
            ).get_auth_service()

    @staticmethod
    def can_provide() -> bool:
        with CredentialsFileManager() as cred_file:
            return cred_file.exists()

    @property
    def token(self) -> Token:
        with TOKEN_REFRESH_LOCK:
            if not self._token.is_going_to_be_expired():
                return self._token

            logger.info("Refreshing access token")
            with self._cred_file:
                new_cred_file_content = self._cred_file.read()
                new_token = new_cred_file_content.to_token()
                new_tracking_uri = new_cred_file_content.host

                if new_cred_file_content == self._last_cred_file_content:
                    self._token = self._auth_service.refresh_token(self._token)
                    self._last_cred_file_content = CredentialsFileContent(
                        host=self._tracking_uri,
                        access_token=self._token.access_token,
                        refresh_token=self._token.refresh_token,
                    )
                    self._cred_file.write(self._last_cred_file_content)
                    return self._token

                if (
                    new_tracking_uri == self._tracking_uri
                    and new_token.to_user_info() == self._token.to_user_info()
                ):
                    self._last_cred_file_content = new_cred_file_content
                    self._token = new_token
                    # recursive
                    return self.token

                raise MlFoundryException(
                    "Credentials on disk changed while mlfoundry was running."
                )

    @property
    def tracking_uri(self) -> str:
        return self._tracking_uri


SESSION_LOCK = threading.RLock()


class ActiveRuns:
    def __init__(self):
        self._active_runs: Dict[str, weakref.ReferenceType[MlFoundryRun]] = dict()

    def add_run(self, run: MlFoundryRun):
        with SESSION_LOCK:
            self._active_runs[run.run_id] = weakref.ref(run)

    def remove_run(self, run: MlFoundryRun):
        with SESSION_LOCK:
            if run.run_id in self._active_runs:
                del self._active_runs[run.run_id]

    def close_active_runs(self):
        with SESSION_LOCK:
            for run_ref in list(self._active_runs.values()):
                run = run_ref()
                if run and run.auto_end:
                    run.end()
            self._active_runs.clear()


ACTIVE_RUNS = ActiveRuns()
atexit.register(ACTIVE_RUNS.close_active_runs)


class Session:
    def __init__(self, cred_provider: CredentialProvider):
        # Note: Whenever a new session is initialized all the active runs are ended
        self._closed = False
        self._cred_provider: CredentialProvider = cred_provider
        self._user_info: UserInfo = self._cred_provider.token.to_user_info()

    def close(self):
        logger.debug("Closing existing session")
        self._closed = True
        self._user_info = None
        self._cred_provider = None

    def _assert_not_closed(self):
        if self._closed:
            raise MlFoundryException(
                "This session has been deactivated.\n"
                "At a time only one `client` (received from "
                "`mlfoundry.get_client()` function call) can be used"
            )

    @property
    def token(self) -> Token:
        return self._cred_provider.token

    @property
    def user_info(self) -> UserInfo:
        self._assert_not_closed()
        return self._user_info

    @property
    def tracking_uri(self) -> str:
        return self._cred_provider.tracking_uri

    def __eq__(self, other: Session) -> bool:
        return (
            type(self._cred_provider) == type(other._cred_provider)
            and self.user_info == other.user_info
            and self.tracking_uri == other.tracking_uri
        )

    def get_mlfoundry_host_creds_builder(self) -> Callable[[], MlflowHostCreds]:
        tracking_uri = append_path_to_rest_tracking_uri(
            self._cred_provider.tracking_uri
        )

        def builder():
            self._assert_not_closed()
            return MlflowHostCreds(
                host=tracking_uri, token=self._cred_provider.token.access_token
            )

        return builder

    def get_monitoring_foundry_host_creds_builder(
        self,
    ) -> Callable[[], MlflowHostCreds]:
        tracking_uri = append_monitoring_path_to_rest_tracking_uri(
            self._cred_provider.tracking_uri
        )

        def builder():
            self._assert_not_closed()
            return MlflowHostCreds(
                host=tracking_uri, token=self._cred_provider.token.access_token
            )

        return builder


ACTIVE_SESSION: Optional[Session] = None


def get_active_session() -> Session:
    return ACTIVE_SESSION


def init_session() -> Session:
    with SESSION_LOCK:
        final_cred_provider = None
        for cred_provider in [EnvCredentialProvider, FileCredentialProvider]:
            if cred_provider.can_provide():
                final_cred_provider = cred_provider()
                break
        if final_cred_provider is None:
            raise MlFoundryException(
                "Please login using `mlfoundry login` command "
                "or `mlfoundry.login()` function call"
            )
        new_session = Session(cred_provider=final_cred_provider)

        global ACTIVE_SESSION
        if ACTIVE_SESSION and ACTIVE_SESSION == new_session:
            return ACTIVE_SESSION

        ACTIVE_RUNS.close_active_runs()

        if ACTIVE_SESSION:
            ACTIVE_SESSION.close()
        ACTIVE_SESSION = new_session

        host_creds_builder = ACTIVE_SESSION.get_mlfoundry_host_creds_builder()

        tfy_rest_store = TruefoundryRestStore(get_host_creds=host_creds_builder)
        artifact_repository = partial(
            TruefoundryArtifactRepository, rest_store=tfy_rest_store
        )
        _artifact_repository_registry.register("s3", artifact_repository)
        _artifact_repository_registry.register("gs", artifact_repository)
        _artifact_repository_registry.register("wasbs", artifact_repository)

        rest_store = RestStore(get_host_creds=host_creds_builder)

        def _get_rest_store(*args, **kwargs):
            return rest_store

        for scheme in ["http", "https"]:
            _tracking_store_registry.register(scheme, _get_rest_store)

        mlflow.set_tracking_uri(ACTIVE_SESSION.tracking_uri)
        logger.info(
            "Logged in to %r as %r (%s)",
            ACTIVE_SESSION.tracking_uri,
            ACTIVE_SESSION.user_info.user_id,
            ACTIVE_SESSION.user_info.email or ACTIVE_SESSION.user_info.user_type.value,
        )
        return ACTIVE_SESSION



import click

from mlfoundry import get_client


@click.group(help="Download files of artifact/model logged with Mlfoundry")
def download():
    ...


@download.command(short_help="Download files of logged model")
@click.option(
    "--fqn",
    required=True,
    type=str,
    help="fqn of the model version",
)
@click.option(
    "--path",
    type=click.Path(file_okay=False, dir_okay=True, exists=True),
    default="./",
    show_default=True,
    help="path where the model files will be downloaded",
)
@click.option(
    "--overwrite",
    is_flag=True,
    default=False,
    help="flag to overwrite any existing files in the `path` directory",
)
def model(fqn: str, path: str, overwrite: bool):
    """
    Download the files of logged model.\n
    """
    client = get_client()
    model_version = client.get_model_version_by_fqn(fqn=fqn)
    download_path = model_version.download(path=path, overwrite=overwrite)
    print(f"Downloaded model files to {download_path}")


@download.command(short_help="Download files of logged artifact")
@click.option(
    "--fqn",
    required=True,
    type=str,
    help="fqn of the artifact version",
)
@click.option(
    "--path",
    type=click.Path(file_okay=False, dir_okay=True, exists=True),
    default="./",
    show_default=True,
    help="path where the artifact files will be downloaded",
)
@click.option(
    "--overwrite",
    is_flag=True,
    default=False,
    help="flag to overwrite any existing files in the `path` directory",
)
def artifact(fqn: str, path: str, overwrite: bool):
    """
    Download the files of logged artifact.\n
    """
    client = get_client()
    artifact_version = client.get_artifact_version_by_fqn(fqn=fqn)
    download_path = artifact_version.download(path=path, overwrite=overwrite)
    print(f"Downloaded artifact files to {download_path}")



"""
This code may not be needed going forward in future.


mlflow.set_experiment("my_experiment")
with mlflow.start_run(run_name=str(time.time())) as run:
    mlflow.set_tag("foo", "bar")

The above snippet adds commit sha in
mlflow.source.git.commit tag
https://www.mlflow.org/docs/latest/tracking.html#system-tags

This does not work if we use MlflowClient directly.
"""


import typing

from git.exc import InvalidGitRepositoryError

from mlfoundry.logger import logger


class GitInfo:
    def __init__(self, path: str):
        try:
            self.repo = self.build_repo(path)
        except InvalidGitRepositoryError as ex:
            # NOTE: gitpython library does not set proper exception message while raising the
            # exception. So we are catching and raising the same exception with proper message
            raise InvalidGitRepositoryError("git repository is not present") from ex

    def build_repo(self, path: str):
        # https://github.com/gitpython-developers/GitPython/blob/cd29f07b2efda24bdc690626ed557590289d11a6/git/cmd.py#L365
        # looks like the import itself may fail in case the git executable
        # is not found
        # putting the import here so that the caller can handle the exception
        import git

        repo = git.Repo(path, search_parent_directories=True)

        return repo

    @property
    def current_commit_sha(self) -> str:
        return self.repo.head.object.hexsha

    @property
    def current_branch_name(self) -> str:
        try:
            branch_name = self.repo.active_branch.name
            return branch_name
        except TypeError as ex:
            # NOTE: TypeError will be raised here if
            # head is in detached state.
            # git checkout commit_sha
            # in this case returning empty string
            logger.warning(f"cannot get branch name because of {ex}")
            return ""

    @property
    def remote_url(self) -> typing.Optional[str]:
        remotes = self.repo.remotes
        if len(remotes) != 1:
            logger.warning("either more than one or no remote detected")
            return None
        return remotes[0].url

    @property
    def diff_patch(self) -> str:
        return self.repo.git.diff("--patch", "HEAD")

    @property
    def is_dirty(self) -> bool:
        return self.repo.is_dirty()
